#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **typing module** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.module.lib.utiltyping` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ attr : typing              }....................
def test_import_typing_attr() -> None:
    '''
    Test the :func:`beartype._util.module.lib.utiltyping.import_typing_attr`
    importer.
    '''

    # ....................{ IMPORTS                            }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilModuleException
    from beartype._util.module.lib.utiltyping import import_typing_attr
    from beartype._util.module.utilmodimport import import_module_attr_or_sentinel
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_9
    from beartype._util.utilobject import SENTINEL
    from pytest import raises
    from typing import Union

    # ....................{ PASS                           }....................
    # Assert that dynamically importing "typing.Union" attribute (guaranteed to
    # be importable under all supported Python versions) is that attribute.
    assert import_typing_attr('Union') is Union

    # ....................{ FAIL                           }....................
    # Assert that attempting to dynamically import a ridiculous attribute
    # (guaranteed to be unimportable under all supported Python versions, we
    # swear) raises the expected exception.
    with raises(_BeartypeUtilModuleException):
        import_typing_attr('FromMyWingsAreShakenTheDewsThatWaken')

    # ....................{ VERSION                        }....................
    #FIXME: On retiring Python 3.8, refactor everything below to import a newer
    #"typing" attribute introduced with a bleeding-edge Python version.
    # If the active Python interpreter targets Python >= 3.9, the
    # "typing.Annotated" attribute is guaranteed to exist. In this case...
    if IS_PYTHON_AT_LEAST_3_9:
        # Defer Python version-specific imports.
        from typing import Annotated

        # Assert that dynamically importing the "typing.Annotated" attribute
        # (guaranteed to be importable under this Python version) is that
        # attribute.
        assert import_typing_attr('Annotated') is Annotated
    # Else, the active Python interpreter targets Python < 3.9 and the
    # "typing.Annotated" attribute is guaranteed to *NOT* exist. In this
    # case...
    else:
        # The "typing_extensions.Annotated" attribute if the third-party
        # "typing_extensions" module is both installed and declares that
        # attribute *OR* "None" otherwise.
        typing_extensions_annotated = import_module_attr_or_sentinel(
            'typing_extensions.Annotated')

        # If that attribute exists...
        if typing_extensions_annotated is not SENTINEL:
            # Assert that dynamically importing the
            # "typing_extensions.Annotated" attribute (guaranteed to be
            # importable by this condition) is that attribute.
            assert import_typing_attr('Annotated') is (
                typing_extensions_annotated)
        # Else, safely reduce to a noop. This should typically *NEVER* happen,
        # as all sane versions of that module declare that attribute. *shrug*



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype configuration unit tests.**

This submodule unit tests the subset of the public API of the :mod:`beartype`
package defined by the private :mod:`beartype._conf.confenum` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_conf_strategy() -> None:
    '''
    Test the public :func:`beartype.BeartypeStrategy` enumeration.
    '''

    # Defer test-specific imports.
    from beartype import BeartypeStrategy

    # Assert this enumeration declares the expected members.
    assert isinstance(BeartypeStrategy.O0, BeartypeStrategy)
    assert isinstance(BeartypeStrategy.O1, BeartypeStrategy)
    assert isinstance(BeartypeStrategy.Ologn, BeartypeStrategy)
    assert isinstance(BeartypeStrategy.On, BeartypeStrategy)


def test_conf_violation_verbosity() -> None:
    '''
    Test the public :func:`beartype.BeartypeViolationVerbosity` enumeration.
    '''

    # Defer test-specific imports.
    from beartype import BeartypeViolationVerbosity

    # Assert this enumeration declares the expected members.
    assert isinstance(BeartypeViolationVerbosity.MINIMAL, int)
    assert isinstance(BeartypeViolationVerbosity.DEFAULT, int)
    assert isinstance(BeartypeViolationVerbosity.MAXIMAL, int)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator exception unit tests.**

This submodule unit tests the :func:`beartype.beartype` decorator with respect
to exceptions raised by the :mod:`beartype._check.error.errorget` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_decor_violation_culprits() -> None:
    '''
    Test the :func:`beartype.beartype` decorator with respect to **type-checking
    violations** (i.e., :exc:`beartype.roar.BeartypeCallHintViolation`
    exceptions raised by callables decorated by that decorator).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintViolation
    from beartype.typing import List
    from pytest import raises

    # ....................{ ASSERTS ~ strong ref           }....................
    class SpiritBearIGiveYouSalmonToGoAway(object):
        '''
        Arbitrary user-defined class.
        '''

        pass

    class SpiritBearIGiftYouHoneyNotToStay(object):
        '''
        Arbitrary user-defined class.
        '''

        pass

    # Arbitrary object guaranteed to *NOT* be an instance of this class.
    #
    # Since CPython permits weak references to almost all instances of
    # user-defined classes (including both the above class and the "object"
    # superclass), this object exhibits the expected behaviour from the
    # "culprits" API exercised below that depends upon weak references.
    SPIRIT_BEAR_REFUSE_TO_GO_AWAY = SpiritBearIGiftYouHoneyNotToStay()

    @beartype
    def when_spirit_bear_hibernates_in_your_bed(
        best_bear_den: SpiritBearIGiveYouSalmonToGoAway) -> None:
        '''
        Arbitrary function decorated by :func:`beartype.beartype` annotated by a
        type hint inducing edge-case behaviour in type-checking violations
        raised by this callable.
        '''

        pass

    # Assert that calling this function with a parameter violating its type hint
    # raises the expected exception.
    with raises(BeartypeCallHintViolation) as exception_info:
        when_spirit_bear_hibernates_in_your_bed(SPIRIT_BEAR_REFUSE_TO_GO_AWAY)

    # Exception captured by the prior call to this wrapper function.
    exception = exception_info.value

    # Assert that exactly one culprit was responsible for this exception.
    assert len(exception.culprits) == 1

    # Culprit of this exception.
    culprit = exception.culprits[0]

    # Assert that this culprit is the same object passed to this function.
    assert culprit is SPIRIT_BEAR_REFUSE_TO_GO_AWAY

    # ....................{ ASSERTS ~ weak ref             }....................
    # List of lists of byte strings violating type hints annotating the
    # we_are_all_spirit_bear() function defined below.
    #
    # Since CPython prohibits weak references to both lists *AND* byte strings,
    # this data structure exhibits known flaws in the "culprits" API exercised
    # below that depends upon weak references.
    YOU_WENT_THERE_SPIRIT_BEAR = (
        b'Why do you sleep in my pinball room, Spirit Bear?')
    DO_NOT_GO_THERE_SPIRIT_BEAR = [[YOU_WENT_THERE_SPIRIT_BEAR]]

    @beartype
    def we_are_all_spirit_bear(
        best_bear_dens: List[List[str]]) -> None:
        '''
        Arbitrary function decorated by :func:`beartype.beartype` annotated by a
        type hint inducing edge-case behaviour in type-checking violations
        raised by this callable.
        '''

        pass

    # Assert that calling this function with a parameter violating its type hint
    # raises the expected exception.
    with raises(BeartypeCallHintViolation) as exception_info:
        we_are_all_spirit_bear(DO_NOT_GO_THERE_SPIRIT_BEAR)

    # Exception captured by the prior call to this wrapper function.
    exception = exception_info.value

    # Assert that exactly two culprits were responsible for this exception.
    assert len(exception.culprits) == 2

    # Root and leaf culprits of this exception.
    root_culprit = exception.culprits[0]
    leaf_culprit = exception.culprits[1]

    # Assert that both of these culprits are the machine-readable
    # representations of their underlying objects rather than those objects.
    # Why? Because CPython prohibits weak references to those objects. *sigh*
    assert isinstance(root_culprit, str)
    assert isinstance(leaf_culprit, str)
    assert root_culprit == repr(DO_NOT_GO_THERE_SPIRIT_BEAR)
    assert leaf_culprit == repr(YOU_WENT_THERE_SPIRIT_BEAR)


def test_decor_violation_types() -> None:
    '''
    Test the
    :func:`beartype.beartype` decorator with respect to the
    :attr:`beartype.BeartypeConf.violation_param_type` and
    :attr:`beartype.BeartypeConf.violation_return_type` configuration
    parameters.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype import (
        BeartypeConf,
        beartype,
    )
    from beartype.typing import (
        List,
        Tuple,
        Union,
    )
    from pytest import (
        raises,
        warns,
    )

    # ..................{ CLASSES                            }..................
    class TheVacantWoods(Exception):
        '''
        Arbitrary exception subclass.
        '''

        pass


    class SpreadRoundHim(Exception):
        '''
        Arbitrary exception subclass.
        '''

        pass


    class LowInTheWest(UserWarning):
        '''
        Arbitrary warning subclass.
        '''

        pass


    class WhereHeStood(UserWarning):
        '''
        Arbitrary warning subclass.
        '''

        pass

    # ..................{ CALLABLES                          }..................
    def the_clear_and(
        garish_hills: List[str], the_distinct_valley) -> (
        Union[int, Tuple[str, ...]]):
        '''
        Arbitrary callable to be decorated by the :func:`beartype.beartype`
        decorator, configured in various ways below.
        '''

        return the_distinct_valley

    # Beartype decorator configured to raise different types of user-defined
    # exceptions on invalid parameters and returns.
    beartype_raise = beartype(conf=BeartypeConf(
        violation_param_type=TheVacantWoods,
        violation_return_type=SpreadRoundHim,
    ))

    # Beartype decorator configured to emit different types of user-defined
    # warnings on invalid parameters and returns.
    beartype_warn = beartype(conf=BeartypeConf(
        violation_param_type=LowInTheWest,
        violation_return_type=WhereHeStood,
        # is_debug=True,
    ))

    # Above callable decorated with type-checking raising custom exceptions.
    the_clear_and_raise = beartype_raise(the_clear_and)

    # Above callable decorated with type-checking emitting custom warnings.
    the_clear_and_warn = beartype_warn(the_clear_and)

    # ..................{ CONSTANTS                          }..................
    # Arbitrary list of strings to be passed to these callables.
    WHITHER_HAVE_FLED = [
        'Low in the west, the clear and garish hills,',
        'The distinct valley and the vacant woods,',
    ]

    # Arbitrary tuple of strings to be passed to these callables.
    THE_HUES_OF_HEAVEN = (
        'Spread round him where he stood. Whither have fled',
        'The hues of heaven that canopied his bower',
    )

    # ..................{ PASS                               }..................
    # Assert that these decorated callables neither raise exceptions nor emit
    # warnings when passed valid parameters.
    assert the_clear_and_raise(WHITHER_HAVE_FLED, THE_HUES_OF_HEAVEN) is (
        THE_HUES_OF_HEAVEN)
    assert the_clear_and_warn(WHITHER_HAVE_FLED, THE_HUES_OF_HEAVEN) is (
        THE_HUES_OF_HEAVEN)

    # Assert that this decorated callable raises the expected exception that
    # this callable was configured to raise on receiving an invalid parameter.
    with raises(TheVacantWoods):
        the_clear_and_raise(THE_HUES_OF_HEAVEN, WHITHER_HAVE_FLED)

    # Assert that this decorated callable raises the expected exception that
    # this callable was configured to raise on returning an invalid return.
    with raises(SpreadRoundHim):
        the_clear_and_raise(WHITHER_HAVE_FLED, WHITHER_HAVE_FLED)

    # Assert that this decorated callable emits the expected warning that
    # this callable was configured to emit on receiving an invalid parameter.
    with warns(LowInTheWest):
        the_clear_and_warn(THE_HUES_OF_HEAVEN, THE_HUES_OF_HEAVEN)

    # Assert that this decorated callable emits the expected warning that
    # this callable was configured to emit on returning an invalid return.
    with warns(WhereHeStood):
        the_clear_and_warn(WHITHER_HAVE_FLED, WHITHER_HAVE_FLED)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype type hints data-driven testing submodule.**

This submodule predefines low-level global constants exercising known edge
cases on behalf of higher-level unit test submodules -- including PEP-compliant
type hints, PEP-noncompliant type hint, and objects satisfying neither.
'''

# ....................{ IMPORTS                            }....................
from beartype._cave._cavefast import (
    AnyType,
    NoneType,
)
from beartype._cave._cavemap import NoneTypeOr
from pytest import fixture

# ....................{ FIXTURES                           }....................
@fixture(scope='session')
def hints_meta(hints_pep_meta, hints_nonpep_meta) -> (
    'Tuple[HintNonpepMetadata]'):
    '''
    Session-scoped fixture yielding a tuple of **PEP-agnostic type hint
    metadata** (i.e.,
    :class:`beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintNonpepMetadata`
    instances, each describing a sample type hint exercising an edge case in the
    :mod:`beartype` codebase -- including both PEP-compliant and -noncompliant
    type hints).

    Parameters
    ----------
    hints_pep_meta : Tuple[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        Tuple of PEP-compliant type hint metadata describing PEP-compliant type
        hints exercising edge cases in the :mod:`beartype` codebase.
    hints_nonpep_meta : Tuple[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintNonpepMetadata]
        Tuple of PEP-noncompliant type hint metadata describing PEP-noncompliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # One world. One-liner. Let's get together and code alright.
    yield hints_pep_meta + hints_nonpep_meta


@fixture(scope='session')
def hints_ignorable(
    hints_pep_ignorable_shallow,
    hints_pep_ignorable_deep,
) -> frozenset:
    '''
    Session-scoped fixture yielding a frozen set of **ignorable
    PEP-agnostic type hints,** including both PEP-compliant *and* -noncompliant
    and shallowly *and* deeply ignorable type hints.

    Parameters
    ----------
    hints_pep_ignorable_shallow : frozenset
        Frozen set of all shallowly ignorable PEP-compliant type hints.
    hints_pep_ignorable_deep : frozenset
        Frozen set of all deeply ignorable PEP-compliant type hints.
    '''

    # I code the one-liner, but I did not code the two-liner.
    yield (
        hints_pep_ignorable_shallow |
        hints_pep_ignorable_deep
    )

# ....................{ FIXTURES ~ not : sets              }....................
@fixture(scope='session')
def not_hints_nonpep(hints_pep_hashable) -> frozenset:
    '''
    Session-scoped fixture yielding a frozen set of various objects that are
    *not* PEP-noncompliant type hints exercising well-known edge cases.

    Parameters
    ----------
    hints_pep_hashable : frozenset
        Tuple of PEP-compliant type hint metadata describing PEP-compliant type
        hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    yield frozenset((
        # Set comprehension of tuples containing PEP-compliant type hints. Although
        # tuples containing PEP-noncompliant type hints are themselves valid
        # PEP-noncompliant type hints supported by @beartype, tuples containing
        # PEP-compliant type hints are invalid and thus unsupported.
        {
            # Tuple containing this PEP-compliant type hint...
            (int, hint_pep_hashable, NoneType,)
            # For each hashable PEP-compliant type hint...
            for hint_pep_hashable in hints_pep_hashable
            # That is neither:
            # * An isinstanceable class.
            # * A string-based forward reference.
            #
            # Both are unique edge cases supported as both PEP 484-compliant outside
            # tuples *AND* beartype-specific inside tuples. Including these hints
            # here would erroneously cause tests to treat tuples containing these
            # hints as *NOT* tuple type hints.
            if not isinstance(hint_pep_hashable, (str, type))
        } |
        # Set comprehension of hashable PEP-compliant non-class type hints.
        hints_pep_hashable |
        # Hashable objects invalid as type hints (e.g., scalars).
        NOT_HINTS_HASHABLE
    ))

# ....................{ NON-HINTS ~ sets                   }....................
#FIXME: Refactor *ALL* remaining globals below into comparable fixtures, please.
NOT_HINTS_HASHABLE = frozenset((
    # Scalar that is neither a type nor string (i.e., forward reference).
    0.12345678910111213141516,
    # Empty tuple.
    (),
    # Tuple containing a scalar that is neither a type nor string.
    (list, 'list', 0xFEEDFACE, NoneType,),
))
'''
Frozen set of various objects that are hashable but nonetheless unsupported by
the :func:`beartype.beartype` decorator as valid type hints.
'''

# ....................{ NON-HINTS ~ tuples                 }....................
NOT_HINTS_UNHASHABLE = (
    # Dictionary.
    {'For all things turn to barrenness':
     'In the dim glass the demons hold,',},
    # List.
    ['The glass of outer weariness,',
     'Made when God slept in times of old.',],
    # Set.
    {'There, through the broken branches, go',
     'The ravens of unresting thought;',},
)
'''
Tuple of various objects that are unhashable and thus unsupported by the
:func:`beartype.beartype` decorator as valid type hints.

Since sets *cannot* by design contain unhashable objects, this container is
defined as a tuple rather than a set.
'''


NOT_HINTS = tuple(NOT_HINTS_HASHABLE) + NOT_HINTS_UNHASHABLE
'''
Tuple of various objects unsupported by the :func:`beartype.beartype` decorator
as valid type hints, including both hashable and unhashable unsupported
objects.

Since sets *cannot* by design contain unhashable objects, this container is
defined as a tuple rather than a set.
'''

# ....................{ NON-PEP ~ classes                  }....................
class NonpepCustom(str):
    '''
    PEP-noncompliant user-defined class subclassing an arbitrary superclass.
    '''

    pass


class NonpepCustomFakeTyping(object):
    '''
    PEP-noncompliant user-defined class subclassing an arbitrary superclass
    erroneously masquerading as a :mod:`typing` class.

    Specifically, this class:

    * Defines the :meth:`__repr__` dunder method to return a string prefixed by
      the substring ``"typing."``.
    * Defines the :attr:`__module__` dunder attribute to be the fully-qualified
      module name ``"typing"``.
    '''

    def __repr__(self) -> str:
        return 'typing.FakeTypingType'
NonpepCustomFakeTyping.__module__ = 'typing'

# ....................{ NON-PEP ~ sets                     }....................
# Note that we intentionally omit the "NonpepCustomFakeTyping" class here, as
# that class masquerades too well as a "typing" class -- so well, in fact, that
# "beartype._util.hint.pep.utilpeptest" functions are incapable of reasonably
# distinguishing instances of that class from actual "typing" type hints.
HINTS_NONPEP = frozenset((
    # Builtin container type.
    list,
    # Builtin scalar type.
    str,
    # User-defined type.
    NonpepCustom,
    # Non-empty tuple containing two types.
    NoneTypeOr[AnyType],
    # Non-empty tuple containing two types and a fully-qualified forward
    # reference.
    (int, 'beartype._cave.._cavefast.NoneType', set)
))
'''
Frozen set of PEP-noncompliant type hints exercising well-known edge cases.
'''


HINTS_NONPEP_UNIGNORABLE = (
    # PEP-noncompliant type.
    str,
    # PEP-noncompliant tuple of types.
    NoneTypeOr[AnyType],
)
'''
Frozen set of **unignorable PEP-noncompliant type hints** (i.e.,
PEP-noncompliant type hints that are *not* ignorable).
'''

# ....................{ NOT ~ sets                         }....................
NOT_HINTS_PEP = (
    # PEP-noncompliant type hints.
    HINTS_NONPEP |
    # Hashable objects invalid as type hints (e.g., scalars).
    NOT_HINTS_HASHABLE
)
'''
Tuple of various objects that are *not* PEP-compliant type hints exercising
well-known edge cases.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **configuration class hierarchy** (i.e., public dataclasses enabling
users to configure :mod:`beartype` with optional runtime behaviours).

Most of the public attributes defined by this private submodule are explicitly
exported to external users in our top-level :mod:`beartype.__init__` submodule.
This private submodule is *not* intended for direct importation by downstream
callers.
'''

# ....................{ TODO                               }....................
#FIXME: Generalize "warning_cls_on_decorator_exception", please. Specifically:
#* Deprecate "warning_cls_on_decorator_exception".
#* Define a new "decoration_exception_type: Optional[TypeException] = None"
#  parameter accepting *ANY* arbitrary exception rather than merely a warning.

#FIXME: [DOCOS] Document all newly defined configuration parameters in our
#reST-formatted docos, please -- including:
#* "claw_is_pep526".
#* "hint_overrides".
#* "violation_door_type".
#* "violation_param_type".
#* "violation_return_type".
#* "violation_type".
#* "violation_verbosity".
#* "warning_cls_on_decorator_exception".

# ....................{ IMPORTS                            }....................
from beartype.roar._roarwarn import (
    _BeartypeConfReduceDecoratorExceptionToWarningDefault,
)
from beartype.typing import (
    TYPE_CHECKING,
    Dict,
    Optional,
)
from beartype._conf.confenum import (
    BeartypeStrategy,
    BeartypeViolationVerbosity,
)
from beartype._conf.confoverrides import (
    BEARTYPE_HINT_OVERRIDES_EMPTY,
    BeartypeHintOverrides,
)
from beartype._conf.conftest import (
    default_conf_kwargs_after,
    default_conf_kwargs_before,
    die_if_conf_kwargs_invalid,
)
from beartype._conf._confget import get_is_color
from beartype._data.hint.datahinttyping import (
    BoolTristateUnpassable,
    DictStrToAny,
    TypeException,
    TypeWarning,
)
from beartype._data.func.datafuncarg import ARG_VALUE_UNPASSED
from beartype._util.utilobject import get_object_type_basename
from threading import Lock

# ....................{ CLASSES                            }....................
class BeartypeConf(object):
    '''
    **Beartype configuration** (i.e., self-caching dataclass encapsulating all
    flags, options, settings, and other metadata configuring each type-checking
    operation performed by :mod:`beartype` -- including each decoration of a
    callable or class by the :func:`beartype.beartype` decorator).

    Attributes
    ----------
    _claw_is_pep526 : bool
        :data:`True` only if type-checking **annotated variable assignments**
        (i.e., :pep:`526`-compliant assignments to local, global, class, and
        instance variables annotated by type hints) when importing modules
        under import hooks published by the :mod:`beartype.claw` subpackage. See
        also the :meth:`__new__` method docstring.
    _conf_args : tuple
        Tuple of the values of *all* possible keyword parameters (in arbitrary
        order) configuring this configuration.
    _conf_kwargs : Dict[str, object]
        Dictionary mapping from the names to values of *all* possible keyword
        parameters configuring this configuration.
    _hash : int
        Precomputed configuration hash returned by the :meth:`__hash__` dunder
        method for efficiency.
    _hint_overrides : Dict
        **Type hint overrides** (i.e., frozen dictionary mapping from arbitrary
        source to target type hints), enabling callers to lie to both their
        users and all other packages other than :mod:`beartype`. This dictionary
        enables callers to externally present a public API annotated by
        simplified type hints while internally instructing :mod:`beartype` to
        privately type-check that API under a completely different set of
        (typically more complicated) type hints.
    _is_color : Optional[bool]
        Tri-state boolean governing how and whether beartype colours
        **type-checking violations** (i.e.,
        :class:`beartype.roar.BeartypeCallHintViolation` exceptions) with
        POSIX-compliant ANSI escape sequences for readability. Specifically, if
        this boolean is:

        * :data:`False`, beartype *never* colours type-checking violations
          raised by callables configured with this configuration.
        * :data:`True`, beartype *always* colours type-checking violations
          raised by callables configured with this configuration.
        * :data:`None`, beartype conditionally colours type-checking violations
          raised by callables configured with this configuration only when
          standard output is attached to an interactive terminal.
    _is_debug : bool
        :data:`True` only if debugging :mod:`beartype`. See also the
        :meth:`__new__` method docstring.
    _is_pep484_tower : bool
        :data:`True` only if enabling support for the :pep:`484`-compliant
        implicit numeric tower. See also the :meth:`__new__` method docstring.
    _is_violation_door_warn : bool
        :data:`True` only if :attr:`violation_door_type` is a warning subclass.
        Note that this is stored only as a negligible optimization to avoid
        needless recomputation of this boolean during code generation.
    _is_violation_param_warn : bool
        :data:`True` only if :attr:`violation_param_type` is a warning subclass.
        Note that this is stored only as a negligible optimization to avoid
        needless recomputation of this boolean during code generation.
    _is_violation_return_warn : bool
        :data:`True` only if :attr:`violation_return_type` is a warning
        subclass. Note that this is stored only as a negligible optimization to
        avoid needless recomputation of this boolean during code generation.
    _is_warning_cls_on_decorator_exception_set : bool
        :data:`True` only if the caller explicitly passed the
        :attr:`_warning_cls_on_decorator_exception` parameter. See
        also the :meth:`__new__` method docstring.
    _repr : Optional[str]
        Either:

        * If the :func:`repr` builtin has yet to call the :meth:`__repr__`
          dunder method, :data:`None`.
        * Else, the machine-readable representation of this configuration,
    _strategy : BeartypeStrategy
        **Type-checking strategy** (i.e., :class:`BeartypeStrategy` enumeration
        member) with which to implement all type-checks in the wrapper function
        dynamically generated by the :func:`beartype.beartype` decorator for
        the decorated callable.
    violation_door_type : TypeException
        **DOOR violation type** (i.e., type of exception raised by the
        :func:`beartype.door.die_if_unbearable` type-checker when the object
        passed to that type-checker violates the type hint passed to that
        type-checker). See also the :meth:`__new__` method docstring.
    _violation_param_type : TypeException
        **Parameter violation type** (i.e., type of exception raised by
        callables generated by the :func:`beartype.beartype` decorator when
        those callables receive parameters violating the type hints annotating
        those parameters). See also the :meth:`__new__` method docstring.
    _violation_return_type : TypeException
        **Return violation type** (i.e., type of exception raised by callables
        generated by the :func:`beartype.beartype` decorator when those
        callables return values violating the type hints annotating those
        returns). See also the :meth:`__new__` method docstring.
    _violation_type : Optional[TypeException]
        **Default violation type** (i.e., type of exception to default whichever
        of the ``violation_door_type``, ``violation_param_type``, and
        ``violation_return_type`` exception types are unpassed and thus
        :data:`None`). See also the :meth:`__new__` method docstring.
    _violation_verbosity : BeartypeViolationVerbosity
        **Violation verbosity** (i.e., positive integer in the inclusive range
        ``[1, 5]`` governing the verbosity of exception messages raised by
        type-checking wrappers generated by the :func:`beartype.beartype`
        decorator when either receiving parameters *or* returning values
        violating their annotated type hints). See also the :meth:`__new__`
        method docstring.
    _warning_cls_on_decorator_exception : Optional[TypeWarning]
        Configuration parameter governing whether the :func:`beartype.beartype`
        decorator reduces otherwise fatal exceptions raised at decoration time
        to equivalent non-fatal warnings of this warning category. See also the
        :meth:`__new__` method docstring.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize this slots list with the implementations of:
    # * The __new__() dunder method.
    # * The __eq__() dunder method.
    # * The __hash__() dunder method.
    # * The __repr__() dunder method.
    # CAUTION: Subclasses declaring uniquely subclass-specific instance
    # variables *MUST* additionally slot those variables. Subclasses violating
    # this constraint will be usable but unslotted, which defeats our purposes.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently called
    # cache dunder methods. Slotting has been shown to reduce read and write
    # costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        '_claw_is_pep526',
        '_conf_args',
        '_conf_kwargs',
        '_hash',
        '_hint_overrides',
        '_is_color',
        '_is_debug',
        '_is_pep484_tower',
        '_is_violation_door_warn',
        '_is_violation_param_warn',
        '_is_violation_return_warn',
        '_is_warning_cls_on_decorator_exception_set',
        '_repr',
        '_strategy',
        '_violation_door_type',
        '_violation_param_type',
        '_violation_return_type',
        '_violation_type',
        '_violation_verbosity',
        '_warning_cls_on_decorator_exception',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        _claw_is_pep526: bool
        _conf_args: tuple
        _conf_kwargs: DictStrToAny
        _hash: int
        _hint_overrides: BeartypeHintOverrides
        _is_color: Optional[bool]
        _is_debug: bool
        _is_pep484_tower: bool
        _is_violation_door_warn: bool
        _is_violation_param_warn: bool
        _is_violation_return_warn: bool
        _is_warning_cls_on_decorator_exception_set: bool
        _repr: Optional[str]
        _strategy: BeartypeStrategy
        _violation_door_type: TypeException
        _violation_param_type: TypeException
        _violation_return_type: TypeException
        _violation_type: Optional[TypeException]
        _violation_verbosity: BeartypeViolationVerbosity
        _warning_cls_on_decorator_exception: Optional[TypeWarning]

    # ..................{ INSTANTIATORS                      }..................
    # Note that this __new__() dunder method implements the superset of the
    # functionality typically implemented by the __init__() dunder method. Due
    # to Python instantiation semantics, the __init__() dunder method is
    # intentionally left undefined. Why? Because Python unconditionally invokes
    # __init__() if defined, even when the initialization performed by that
    # __init__() has already been performed for the cached instance returned by
    # __new__(). In short, __init__() and __new__() are largely mutually
    # exclusive; one typically defines one or the other but *NOT* both.

    def __new__(
        cls,

        # Optional keyword-only parameters.
        *,

        # Uncomment us when implementing O(n) type-checking, please.
        # check_time_max_multiplier: Union[int, None] = 1000,
        claw_is_pep526: bool = True,
        hint_overrides: BeartypeHintOverrides = BEARTYPE_HINT_OVERRIDES_EMPTY,
        is_color: BoolTristateUnpassable = ARG_VALUE_UNPASSED,
        is_debug: bool = False,
        is_pep484_tower: bool = False,
        strategy: BeartypeStrategy = BeartypeStrategy.O1,
        violation_door_type: Optional[TypeException] = None,
        violation_param_type: Optional[TypeException] = None,
        violation_return_type: Optional[TypeException] = None,
        violation_type: Optional[TypeException] = None,
        violation_verbosity: BeartypeViolationVerbosity = (
            BeartypeViolationVerbosity.DEFAULT),
        warning_cls_on_decorator_exception: Optional[TypeWarning] = (
            _BeartypeConfReduceDecoratorExceptionToWarningDefault),
    ) -> 'BeartypeConf':
        '''
        Instantiate this configuration if needed (i.e., if *no* prior
        configuration with these same parameters was previously instantiated)
        *or* reuse that previously instantiated configuration otherwise.

        This dunder methods guarantees beartype configurations to be memoized:

        .. code-block:: python

           >>> from beartype import BeartypeConf
           >>> BeartypeConf() is BeartypeConf()
           True

        This memoization is *not* merely an optimization. The
        :func:`beartype.beartype` decorator internally memoizes the private
        closure it creates and returns on the basis of this configuration,
        which *must* thus also be memoized.

        Parameters
        ----------
        check_time_max_multiplier : Union[int, None] = 1000
            **Deadline multiplier** (i.e., positive integer instructing
            :mod:`beartype` to prematurely halt the current type-check when the
            total running time of the active Python interpreter exceeds this
            integer multiplied by the running time consumed by both the current
            type-check and all prior type-checks *and* the caller also passed a
            non-default ``strategy``) *or* ``None`` if :mod:`beartype` should
            never prematurely halt runtime type-checks.

            Increase this quantity to type-check more container items at a cost
            of decreasing application responsiveness. Likewise, decrease this
            quantity to increase application responsiveness at a cost of
            type-checking fewer container items.

            Ignored when ``strategy`` is :attr:`BeartypeStrategy.O1`, as that
            strategy is already effectively instantaneous; imposing deadlines
            and thus bureaucratic bookkeeping on that strategy would only
            reduce its efficiency for no good reason, which is a bad reason.

            Defaults to 1000, in which case a maximum of %0.1 of the total
            runtime of the active Python process will be devoted to performing
            non-constant :mod:`beartype` type-checks over container items. This
            default has been carefully tuned to strike a reasonable balance
            between runtime type-check coverage and application responsiveness,
            typically enabling smaller containers to be fully type-checked
            without noticeably impacting codebase performance.

            *Theory time.* Let:

            * ``T`` be the total time this interpreter has been running.
            * ``b`` be the total time :mod:`beartype` has spent type-checking in
              this interpreter.

            Clearly, ``b <= T``. Generally, ``b <<<<<<< T`` (i.e., type-checks
            consume much less time than the total time consumed by the process).
            However, it's all too easy to exhibit worst-case behaviour of
            ``b ~= T`` (i.e., type-checks consume most of the total time). How?
            By passing the :func:`beartype.door.is_bearable` tester an absurdly
            large nested container subject to the non-default ``strategy`` of
            :attr:`BeartypeStrategy.On`.

            This deadline multiplier mitigates that worst-case behaviour.
            Specifically, :mod:`beartype` will prematurely halt any iterative
            type-check across a container when this constraint is triggered:

            .. code-block:: python

               b * check_time_max_multiplier >= T
        claw_is_pep526 : bool, optional
            :data:`True` only if implicitly type-checking **annotated variable
            assignments** (i.e., :pep:`526`-compliant assignments to local,
            global, class, and instance variables annotated by type hints) when
            importing modules under import hooks published by the
            :mod:`beartype.claw` subpackage by injecting calls to the
            :func:`beartype.door.die_if_unbearable` function immediately *after*
            those assignments in those modules. Enabling this boolean:

            * Effectively augments :mod:`beartype` into a full-blown **hybrid
              runtime-static type-checker** (i.e., performing both standard
              runtime type-checking *and* non-standard static type-checking at
              runtime).
            * Adds negligible runtime overhead to all annotated variable
              assignments in all modules imported under those import hooks.
              Although the *individual* cost of this overhead for any given
              assignment is negligible, the *aggregate* cost across all such
              assignments could be non-negligible in worst-case use cases.

            Ideally, this boolean should only be disabled for a small subset of
            performance-sensitive modules *after* profiling those modules to
            suffer performance regressions under import hooks published by the
            :mod:`beartype.claw` subpackage. Defaults to :data:`True`.
        hint_overrides : BeartypeHintOverrides
            **Type hint overrides** (i.e., frozen dictionary mapping from
            arbitrary source to target type hints), enabling callers to lie to
            both their users and all other packages other than :mod:`beartype`.
            This dictionary enables callers to externally present a public API
            annotated by simplified type hints while internally instructing
            :mod:`beartype` to privately type-check that API under a completely
            different set of (typically more complicated) type hints. Doing so
            preserves a facade of simplicity for downstream consumers like end
            users, static type-checkers, and document generators. Defaults to
            the empty dictionary.

            Specifically, for each source type hint annotating each callable,
            class, or variable assignment observed by :mod:`beartype`, if that
            source type hint is a key of this dictionary, :mod:`beartype` maps
            that source type hint to the corresponding target type hint in this
            dictionary. That target type hint then globally "overrides" (i.e.,
            replaces, substitutes for) that source type hint. :mod:`beartype`
            then uses that target type hint in place of that source type hint.

            For example, consider this Abomination Unto the Eyes of Guido:

            .. code-block:: python

               from beartype, BeartypeConf, BeartypeHintOverrides

               # @beartype decorator configured to expand all "float" type hints
               # to "int | float" type hints.
               lyingbeartype = beartype(conf=BeartypeConf(
                   hint_overrides=BeartypeHintOverrides({float: int | float})))

               # The @lyingbeartype decorator now expands this signature...
               @lyingbeartype
               def lies(all_lies: list[int]) -> int:
                   return all_lies[0]

               # ...as if it had been annotated like this instead.
               @beartype
               def lies(all_lies: list[int | float]) -> int | float:
                   return all_lies[0]
        is_color : BoolTristateUnpassable
            Tri-state boolean governing how and whether beartype colours
            **type-checking violations** (i.e.,
            :class:`beartype.roar.BeartypeCallHintViolation` exceptions) with
            POSIX-compliant ANSI escape sequences for readability. Specifically,
            if this boolean is:

            * :data:`False`, beartype *never* colours type-checking violations
              raised by callables configured with this configuration.
            * :data:`True`, beartype *always* colours type-checking violations
              raised by callables configured with this configuration.
            * :data:`None`, beartype conditionally colours type-checking
              violations raised by callables configured with this configuration
              only when standard output is attached to an interactive terminal.

            The ``${BEARTYPE_IS_COLOR}`` environment variable globally overrides
            *all* attempts by *all* callers to explicitly pass this parameter,
            enabling end users to enforce a global colour policy across their
            full app stack. If ``${BEARTYPE_IS_COLOR}`` is set to a different
            value than that of this parameter, this constructor emits a
            non-fatal :class:`.BeartypeConfShellVarWarning` warning informing
            the caller of this configuration conflict. To avoid this conflict,
            open-source libraries are recommended to *not* pass this parameter;
            ideally, *only* end user apps should pass this parameter.

            Effectively defaults to :data:`None`. Technically, this parameter
            defaults to a private magic constant *not* intended to be passed by
            callers, enabling :mod:`beartype` to reliably detect whether the
            caller has explicitly passed this parameter or not.
        is_debug : bool, optional
            :data:`True` only if debugging :mod:`beartype`. Enabling this
            boolean:

            * Prints the definition (including both the signature and body) of
              each type-checking wrapper function dynamically generated by
              :mod:`beartype` to standard output.
            * Caches the body of each type-checking wrapper function dynamically
              generated by :mod:`beartype` with the standard :mod:`linecache`
              module, enabling these function bodies to be introspected at
              runtime *and* improving the readability of tracebacks whose call
              stacks contain one or more calls to these
              :func:`beartype.beartype`-decorated functions.
            * Appends to the declaration of each **hidden parameter** (i.e.,
              whose name is prefixed by ``"__beartype_"`` and whose value is
              that of an external attribute internally referenced in the body of
              that function) the machine-readable representation of the initial
              value of that parameter, stripped of newlines and truncated to a
              hopefully sensible length. Since the
              :func:`beartype._util.text.utiltextrepr.represent_object` function
              called to do so is shockingly slow, these substrings are
              conditionally embedded in the returned signature *only* when
              enabling this boolean.

            Defaults to :data:`False`.
        is_pep484_tower : bool, optional
            :data:`True` only if enabling support for the :pep:`484`-compliant
            **implicit numeric tower** (i.e., lossy conversion of integers to
            floating-point numbers *and* both integers and floating-point
            numbers to complex numbers). Specifically, enabling this instructs
            :mod:`beartype` to automatically expand:

            * All :class:`float` type hints to ``float | int``, thus implicitly
              accepting both integers and floating-point numbers for objects
              annotated as only accepting floating-point numbers.
            * All :class:`complex` type hints to ``complex | float | int``, thus
              implicitly accepting integers, floating-point, and complex numbers
              for objects annotated as only accepting complex numbers.

            Defaults to :data:`False` to minimize precision error introduced by
            lossy conversions from integers to floating-point numbers to complex
            numbers. Since most integers do *not* have exact representations
            as floating-point numbers, each conversion of an integer into a
            floating-point number typically introduces a small precision error
            that accumulates over multiple conversions and operations into a
            larger precision error. Enabling this improves the usability of
            public APIs at a cost of introducing precision errors.
        strategy : BeartypeStrategy, optional
            **Type-checking strategy** (i.e., :class:`BeartypeStrategy`
            enumeration member) with which to implement all type-checks in the
            wrapper function dynamically generated by the
            :func:`beartype.beartype` decorator for the decorated callable.
            Defaults to :attr: `BeartypeStrategy.O1`, the ``O(1)`` constant-time
            strategy.
        violation_door_type : Optional[TypeException]
            **DOOR violation type** (i.e., type of exception raised by the
            :func:`beartype.door.die_if_unbearable` type-checker when the object
            passed to that type-checker violates the type hint passed to that
            type-checker). Defaults to :data:`None`, in which case this type
            defaults to either:

            * If ``violation_type`` is passed and *not* :data:`None`, that type.
            * Else, :exc:`.BeartypeDoorHintViolation`.
        violation_param_type : Optional[TypeException]
            **Parameter violation type** (i.e., type of exception raised by
            callables generated by the :func:`beartype.beartype` decorator when
            those callables receive parameters violating the type hints
            annotating those parameters). Defaults to :data:`None`, in which
            case this type defaults to either:

            * If ``violation_type`` is passed and *not* :data:`None`, that type.
            * Else, :exc:`.BeartypeCallHintParamViolation`.
        violation_return_type : Optional[TypeException]
            **Return violation type** (i.e., type of exception raised by
            callables generated by the :func:`beartype.beartype` decorator when
            those callables return values violating the type hints annotating
            those returns). Defaults to :data:`None`, in which case this type
            defaults to either:

            * If ``violation_type`` is passed and *not* :data:`None`, that type.
            * Else, :exc:`.BeartypeCallHintReturnViolation`.
        violation_type : Optional[TypeException]
            **Default violation type** (i.e., type of exception to default
            whichever of the ``violation_door_type``, ``violation_param_type``,
            and ``violation_return_type`` exception types are unpassed and thus
            :data:`None`).  This parameter is merely a convenience enabling
            callers to trivially control the ``violation_door_type``,
            ``violation_param_type``, and ``violation_return_type`` parameters
            *without* having to explicitly pass all three of those parameters.
            Specifically, if this parameter is:

            * *Not* :data:`None`, then the ``violation_door_type``,
              ``violation_param_type``, and ``violation_return_type`` parameters
              all default to the value of this parameter.
            * :data:`None`, then:

              * ``violation_door_type`` defaults to
                :exc:`.BeartypeDoorHintViolation`.
              * ``violation_param_type`` defaults to
                :exc:`.BeartypeCallHintParamViolation`.
              * ``violation_return_type`` defaults to
                :exc:`.BeartypeCallHintReturnViolation`.

            Defaults to :data:`None`.
        violation_verbosity : BeartypeViolationVerbosity, optional
            **Violation verbosity** (i.e., positive integer in the inclusive
            range ``[1, 5]`` governing the verbosity of exception messages
            raised by type-checking wrappers generated by the
            :func:`beartype.beartype` decorator when either receiving parameters
            *or* returning values violating their annotated type hints).
            Defaults to :attr:`.BeartypeViolationVerbosity.DEFAULT`.
        warning_cls_on_decorator_exception : Optional[TypeWarning]
            Configuration parameter governing whether the
            :func:`beartype.beartype` decorator reduces what would otherwise be
            fatal exceptions raised at decoration time to equivalent non-fatal
            warnings of the passed **warning category** (i.e., subclass of the
            standard :class:`Warning` class). Specifically, this parameter may
            be either:

            * :data:`None`, in which case the :func:`beartype.beartype`
              decorator raises fatal exceptions at decoration time.
            * A warning category, in which case the :func:`beartype.beartype`
              decorator reduces fatal exceptions to non-fatal warnings of this
              category at decoration time.

            Defaults to a private warning type *not* intended to be passed by
            callers, enabling :mod:`beartype` to reliably detect when the caller
            has *not* explicitly passed this parameter and respond accordingly
            by defaulting this parameter to a context-dependent value. Notably,
            if this parameter is *not* explicitly passed:

            * The :func:`beartype.beartype` decorator defaults this parameter to
              :data:`None`, thus raising decoration-time exceptions.
            * The :mod:`beartype.claw` API defaults this parameter to the public
              :class:`beartype.roar.BeartypeClawDecorWarning` warning category,
              thus reducing decoration-time exceptions to warnings of that
              category when performing import hooks. This default behaviour
              significantly increases the likelihood that import hooks installed
              by :mod:`beartype.claw` will successfully decorate the entirety of
              their target packages rather than prematurely halt with a single
              fatal exception at the first decoration issue.

        Returns
        -------
        BeartypeConf
            Beartype configuration memoized with these parameters.

        Raises
        ------
        BeartypeConfParamException
            If either:

            * ``is_color`` is *not* a tri-state boolean.
            * ``is_debug`` is *not* a boolean.
            * ``is_pep484_tower`` is *not* a boolean.
            * ``strategy`` is *not* a :class:`BeartypeStrategy` enumeration
              member.
            * ``warning_cls_on_decorator_exception`` is neither :data:`None`
              *nor* a **warning category** (i.e., :class:`Warning` subclass).
        BeartypeConfShellVarException
            If either:

            * The external ``${BEARTYPE_IS_COLOR}`` shell environment variable
              is set to an unrecognized string (i.e., neither ``"True"``,
              ``"False"``, nor ``"None"``).
        '''

        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # CAUTION: Synchronize this tuple with the similar "self._conf_kwargs"
        # dictionary defined below.
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # In a non-reentrant thread lock specific to beartype configurations...
        #
        # Note that this lock is potentially overkill and thus unnecessary.
        # Nonetheless, since the number of beartype configurations instantiated
        # over the lifetime of the average Python interpreter is small, since
        # non-reentrant thread locks are reasonably fast to enter, and since the
        # cost of race conditions is high, this lock does no real-world harm and
        # may actually do a great deal of real-world good. Safety first, all!
        with _beartype_conf_lock:
            # ..................{ CACHE                      }..................
            # Validate and possibly override the "is_color" parameter by the
            # value of the ${BEARTYPE_IS_COLOR} environment variable (if set).
            is_color = get_is_color(is_color)

            # Efficiently hashable tuple of these parameters in arbitrary order.
            conf_args = (
                claw_is_pep526,
                hint_overrides,
                is_color,
                is_debug,
                is_pep484_tower,
                strategy,
                violation_door_type,
                violation_param_type,
                violation_return_type,
                violation_type,
                violation_verbosity,
                warning_cls_on_decorator_exception,
            )

            # If this method has already instantiated a configuration with these
            # parameters, return that configuration for consistency and
            # efficiency.
            if conf_args in _beartype_conf_args_to_conf:
                return _beartype_conf_args_to_conf[conf_args]
            # Else, this method has *NOT* yet instantiated a configuration with
            # these parameters. In this case, continue to do so and then cache
            # that configuration.

            # Dictionary mapping from the names to values of *ALL* possible
            # keyword parameters configuring this configuration, intentionally
            # defined *AFTER* this method first attempts to efficiently reduce
            # to a noop by returning a previously instantiated configuration.
            conf_kwargs = dict(
                claw_is_pep526=claw_is_pep526,
                hint_overrides=hint_overrides,
                is_color=is_color,
                is_debug=is_debug,
                is_pep484_tower=is_pep484_tower,
                strategy=strategy,
                violation_door_type=violation_door_type,
                violation_param_type=violation_param_type,
                violation_return_type=violation_return_type,
                violation_type=violation_type,
                violation_verbosity=violation_verbosity,
                warning_cls_on_decorator_exception=(
                    warning_cls_on_decorator_exception),
            )

            # Default all parameters not explicitly passed by the user to sane
            # defaults *BEFORE* validating these parameters.
            default_conf_kwargs_before(conf_kwargs)

            # If one or more passed parameters are invalid, raise an exception.
            die_if_conf_kwargs_invalid(conf_kwargs)
            # Else, all passed parameters are valid.

            # Default all parameters not explicitly passed by the user to sane
            # defaults *AFTER* validating these parameters.
            default_conf_kwargs_after(conf_kwargs)

            # ..................{ INSTANTIATE                }..................
            # Instantiate a new configuration of this type.
            self = super().__new__(cls)

            # Nullify critical instance variables for safety.
            self._repr = None

            # Precompute the hash to be returned by the __hash__() dunder method
            # as the hash of a tuple containing these parameters in an arbitrary
            # (albeit well-defined) order.
            #
            # Note this has been profiled to be the optimal means of hashing
            # object attributes in Python, where "optimal" means:
            # * Optimally fast. CPython in particular optimizes the creation and
            #   garbage collection of "small" tuples, where "small" is
            #   ill-defined but almost certainly applies here.
            # * Optimally uniformly distributed, thus minimizing the likelihood
            #   of expensive hash collisions.
            self._hash = hash(conf_args)

            # Store data structures encapsulating these passed parameters for
            # subsequent reuse *BEFORE* possibly modifying the values of these
            # parameters below.
            self._conf_args = conf_args
            self._conf_kwargs = conf_kwargs

            # Assert that these two data structures encapsulate the same number
            # of configuration parameters (as a feeble safety check).
            assert len(self._conf_args) == len(self._conf_kwargs)

            # Cache this configuration with all relevant dictionary singletons
            # *BEFORE* possibly modifying the values of passed parameters below.
            _beartype_conf_args_to_conf[conf_args] = self

            # ..................{ CLASSIFY                   }..................
            # Classify all passed parameters that have now been possibly
            # modified above with this configuration.
            #
            # Note that this classification intentionally accesses these
            # parameters from the "conf_kwargs" dictionary possibly modified by
            # the above call to the default_conf_kwargs() function rather than
            # the original passed values of these parameters.
            self._claw_is_pep526 = conf_kwargs['claw_is_pep526']  # pyright: ignore
            self._hint_overrides = conf_kwargs['hint_overrides']  # pyright: ignore
            self._is_color = conf_kwargs['is_color']  # pyright: ignore
            self._is_debug = conf_kwargs['is_debug']  # pyright: ignore
            self._is_pep484_tower = conf_kwargs['is_pep484_tower']  # pyright: ignore
            self._strategy = conf_kwargs['strategy']  # pyright: ignore
            self._violation_door_type = conf_kwargs['violation_door_type']  # pyright: ignore
            self._violation_param_type = conf_kwargs['violation_param_type']  # pyright: ignore
            self._violation_return_type = conf_kwargs['violation_return_type']  # pyright: ignore
            self._violation_type = conf_kwargs['violation_type']  # pyright: ignore
            self._violation_verbosity = conf_kwargs['violation_verbosity']  # pyright: ignore

            # Classify all remaining instance variables.
            self._is_violation_door_warn = issubclass(
                self._violation_door_type, Warning)  # pyright: ignore
            self._is_violation_param_warn = issubclass(
                self._violation_param_type, Warning)  # pyright: ignore
            self._is_violation_return_warn = issubclass(
                self._violation_return_type, Warning)  # pyright: ignore

            # ..................{ CLASSIFY ~ more            }..................
            # If the value of the "warning_cls_on_decorator_exception" parameter
            # is still the default private fake warning category established
            # above, then the caller failed to explicitly pass a valid value. In
            # this case...
            if (
                warning_cls_on_decorator_exception is
                _BeartypeConfReduceDecoratorExceptionToWarningDefault
            ):
                # Note this fact for subsequent reference elsewhere (e.g., in
                # the "beartype.claw" subpackage).
                self._is_warning_cls_on_decorator_exception_set = False

                # Default this parameter to "None" for safety. Since this
                # default private fake warning category is *NOT* an actual
                # warning category intended for real-world use, this category
                # *MUST* be replaced with a sane default that is safely usable.
                warning_cls_on_decorator_exception = None
            # Else, the caller explicitly passed a valid value for this
            # parameter. In this case, preserve this value and note this fact.
            else:
                self._is_warning_cls_on_decorator_exception_set = True

            self._warning_cls_on_decorator_exception = (
                warning_cls_on_decorator_exception)

        # Return this configuration.
        return self

    # ..................{ PROPERTIES                         }..................
    # Read-only public properties effectively prohibiting mutation of their
    # underlying private attributes.

    #FIXME: Publicly document this in our reST-formatted docos, please.
    @property
    def kwargs(self) -> DictStrToAny:
        '''
        **Beartype configuration keyword dictionary** (i.e., dictionary mapping
        from the names of all keyword parameters accepted by the :meth:`__new__`
        method to the corresponding values of those parameters in this
        configuration).

        This property can be used to permute new configurations from existing
        configurations, overriding only a small handful of parameters while
        preserving all other parameters as is: e.g.,

        .. code-block:: python

           # Arbitrary input beartype configuration.
           conf = BeartypeConf(is_color=True)

           # Permuted output beartype configuration keyword dictionary.
           kwargs = conf.kwargs.copy()
           kwargs['is_debug'] = True

           # Output beartype configuration permuted from this input.
           debug_conf = BeartypeConf(**kwargs)

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._conf_kwargs

    # ..................{ PROPERTIES ~ options               }..................
    # Read-only public properties with which this configuration was originally
    # instantiated (as keyword-only parameters).

    @property
    def hint_overrides(self) -> BeartypeHintOverrides:
        '''
        **Type hint overrides** (i.e., frozen dictionary mapping from arbitrary
        source to target type hints), enabling callers to lie to both their
        users and all other packages other than :mod:`beartype`. This dictionary
        enables callers to externally present a public API annotated by
        simplified type hints while internally instructing :mod:`beartype` to
        privately type-check that API under a completely different set of
        (typically more complicated) type hints.

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._hint_overrides


    @property
    def strategy(self) -> BeartypeStrategy:
        '''
        **Type-checking strategy** (i.e., :class:`BeartypeStrategy`
        enumeration member) with which to implement all type-checks in the
        wrapper function dynamically generated by the
        :func:`beartype.beartype` decorator for the decorated callable.

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._strategy


    @property
    def warning_cls_on_decorator_exception(self) -> (
        Optional[TypeWarning]):
        '''
        Configuration parameter governing whether the :func:`beartype.beartype`
        decorator reduces otherwise fatal exceptions raised at decoration time
        to equivalent non-fatal warnings of this warning category.

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._warning_cls_on_decorator_exception

    # ..................{ PROPERTIES ~ options : bool        }..................
    # Read-only public properties with which this configuration was originally
    # instantiated (as keyword-only parameters).

    @property
    def claw_is_pep526(self) -> bool:
        '''
        :data:`True` only if type-checking **annotated variable assignments**
        (i.e., :pep:`526`-compliant assignments to local, global, class, and
        instance variables annotated by type hints) when importing modules
        under import hooks published by the :mod:`beartype.claw` subpackage.

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._claw_is_pep526


    @property
    def is_color(self) -> Optional[bool]:
        '''
        Tri-state boolean governing how and whether beartype colours
        **type-checking violations** (i.e.,
        :class:`beartype.roar.BeartypeCallHintViolation` exceptions) with
        POSIX-compliant ANSI escape sequences for readability. Specifically, if
        this boolean is:

        * :data:`False`, beartype *never* colours type-checking violations
          raised by callables configured with this configuration.
        * :data:`True`, beartype *always* colours type-checking violations
          raised by callables configured with this configuration.
        * :data:`None`, beartype conditionally colours type-checking violations
          raised by callables configured with this configuration only when
          standard output is attached to an interactive terminal.

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._is_color


    @property
    def is_debug(self) -> bool:
        '''
        :data:`True` only if debugging :mod:`beartype`.

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._is_debug


    @property
    def is_pep484_tower(self) -> bool:
        '''
        :data:`True` only if enabling support for the :pep:`484`-compliant
        implicit numeric tower.

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._is_pep484_tower

    # ..................{ PROPERTIES ~ options : violation   }..................
    @property
    def violation_door_type(self) -> TypeException:
        '''
        **DOOR violation type** (i.e., type of exception raised by the
        :func:`beartype.door.die_if_unbearable` type-checker when the object
        passed to that type-checker violates the type hint passed to that
        type-checker).

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._violation_door_type


    @property
    def violation_param_type(self) -> TypeException:
        '''
        **Parameter violation type** (i.e., type of exception raised by
        callables generated by the :func:`beartype.beartype` decorator when
        those callables receive parameters violating the type hints annotating
        those parameters).

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._violation_param_type


    @property
    def violation_return_type(self) -> TypeException:
        '''
        **Return violation type** (i.e., type of exception raised by callables
        generated by the :func:`beartype.beartype` decorator when those
        callables return values violating the type hints annotating those
        returns).

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._violation_return_type


    @property
    def violation_type(self) -> Optional[TypeException]:
        '''
        **Default violation type** (i.e., type of exception to default whichever
        of the ``violation_door_type``, ``violation_param_type``, and
        ``violation_return_type`` exception types are unpassed and thus
        :data:`None`).

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._violation_type


    @property
    def violation_verbosity(self) -> BeartypeViolationVerbosity:
        '''
        **Violation verbosity** (i.e., positive integer in the inclusive range
        ``[1, 5]`` governing the verbosity of exception messages raised by
        type-checking wrappers generated by the :func:`beartype.beartype`
        decorator when either receiving parameters *or* returning values
        violating their annotated type hints).

        See Also
        --------
        :meth:`__new__`
            Further details.
        '''

        return self._violation_verbosity

    # ..................{ DUNDERS                            }..................
    def __eq__(self, other: object) -> bool:
        '''
        **Beartype configuration equality comparator.**

        Parameters
        ----------
        other : object
            Arbitrary object to be compared for equality against this
            configuration.

        Returns
        -------
        Union[bool, type(NotImplemented)]
            Either:

            * If this other object is also a beartype configuration, either:

              * If these configurations share the same settings, :data:`True`.
              * Else, :data:`False`.

            * Else, :data:`NotImplemented`.

        See Also
        --------
        :func:`_hash_beartype_conf`
            Further details.
        '''

        # Return either...
        return (
            # If this other object is also a beartype configuration, true only
            # if these configurations share the same settings;
            self._conf_args == other._conf_args
            if isinstance(other, BeartypeConf) else
            # Else, this other object is *NOT* also a beartype configuration. In
            # this case, the standard singleton informing Python that this
            # equality comparator fails to support this comparison.
            NotImplemented  # type: ignore[return-value]
        )


    def __hash__(self) -> int:
        '''
        **Hash** (i.e., non-negative integer quasi-uniquely identifying this
        beartype configuration with respect to hashable container membership).

        Returns
        -------
        int
            Hash of this configuration.
        '''

        # Return the precomputed hash for this configuration.
        return self._hash


    def __repr__(self) -> str:
        '''
        **Beartype configuration representation** (i.e., machine-readable
        string which, when dynamically evaluated as code, restores access to
        this exact configuration object).

        Returns
        -------
        str
            Representation of this configuration.
        '''

        # If machine-readable representation of this configuration has yet to be
        # computed...
        if self._repr is None:
            # Initialize this representation to the unqualified basename of the
            # class of this configuration.
            conf_repr = f'{get_object_type_basename(self)}('

            # Dictionary mapping from the names to values of *ALL* possible
            # keyword parameters configuring the default beartype configuration.
            KWARGS_DEFAULT = BEARTYPE_CONF_DEFAULT._conf_kwargs

            # For the name and value of each keyword parameter with which this
            # configuration was instantiated...
            for kwarg_name, kwarg_value in self._conf_kwargs.items():
                # If this value differs from that of the default value for this
                # keyword parameter...
                if kwarg_value != KWARGS_DEFAULT[kwarg_name]:
                    # Append a comma-delimited representation of this keyword
                    # argument to this representation.
                    conf_repr += f'{kwarg_name}={kwarg_value}, '
                # Else, this value is the default value for this keyword
                # parameter. In this case, silently ignore this value. Appending
                # this value to this representation would convey *NO* meaningful
                # semantics and, indeed, only inhibit the readability of this
                # representation for end users and developers alike.

            # If this representation is suffixed by a whitespaced comma, remove
            # that suffix.
            if conf_repr[-2:] == ', ':
                conf_repr = conf_repr[:-2]
            # Else, this representation is *NOT* suffixed by a comma. In this
            # case, preserve this representation as is.

            # Preserve this representation for subsequent use.
            self._repr = f'{conf_repr})'
        # Else, the machine-readable representation of this configuration has
        # already been computed.

        # Return the machine-readable representation of this configuration.
        return self._repr

# ....................{ PRIVATE ~ globals                  }....................
_beartype_conf_lock = Lock()
'''
**Non-reentrant beartype configuration thread lock** (i.e., low-level thread
locking mechanism implemented as a highly efficient C extension, defined as an
global for non-reentrant reuse elsewhere as a context manager).
'''


_beartype_conf_args_to_conf: Dict[tuple, BeartypeConf] = {}
'''
Non-thread-safe **beartype configuration parameter cache** (i.e., dictionary
mapping from the hash of each set of parameters accepted by a prior call of the
:meth:`BeartypeConf.__new__` instantiator to the unique :class:`BeartypeConf`
instance instantiated by that call).

Caveats
-------
**This cache is non-thread-safe.** However, since this cache is only used as a
memoization optimization, the only harmful consequences of a race condition
between threads contending over this cache is a mildly inefficient (but
otherwise harmless) repeated re-memoization of duplicate configurations.
'''

# ....................{ GLOBALS                            }....................
# This global is intentionally defined *AFTER* all other attributes above, which
# this global implicitly assumes to be defined.
BEARTYPE_CONF_DEFAULT = BeartypeConf()
'''
**Default beartype configuration** (i.e., :class:`BeartypeConf` class
instantiated with *no* parameters and thus default parameters), globalized to
trivially optimize external access to this configuration throughout this
codebase.

Note that this global is *not* publicized to end users, who can simply
instantiate ``BeartypeConf()`` to obtain the same singleton.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype import hook enumerations** (i.e., :class:`enum.Enum` subclasses
enumerating various kinds of divergent strategies and processes specific to
import hooks defined by the :mod:`beartype.claw` subpackage).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
# from beartype.typing import Literal
from enum import (
    Enum,
    auto as next_enum_member_value,
    unique as die_unless_enum_member_values_unique,
)

# ....................{ ENUMS                              }....................
@die_unless_enum_member_values_unique
class BeartypeClawCoverage(Enum):
    '''
    Enumeration of all kinds of **import hook coverage** (i.e., competing
    package scopes over which to apply import hooks defined by the
    :mod:`beartype.claw` subpackage, each with concomitant tradeoffs with
    respect to runtime complexity and quality assurance).

    Attributes
    ----------
    PACKAGES_ALL : EnumMemberType
        **All-packages coverage** (i.e, hooking imports into *all* packages,
        including both third-party packages *and* standard packages bundled with
        Python in the standard library). This coverage is typically applied by a
        caller calling the :func:`beartype.claw.beartype_all` import hook.
    PACKAGES_MANY : EnumMemberType
        **Many-packages coverage** (i.e, hooking imports into two or more
        explicitly specified packages). This coverage is typically applied by a
        caller calling the :func:`beartype.claw.beartype_packages` import hook.
    PACKAGES_ONE : EnumMemberType
        **One-package coverage** (i.e, hooking imports into only one explicitly
        specified package). This coverage is typically applied by a caller
        calling the :func:`beartype.claw.beartype_package` import hook.
    '''

    PACKAGES_ALL = next_enum_member_value()
    PACKAGES_MANY = next_enum_member_value()
    PACKAGES_ONE = next_enum_member_value()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype validator unit tests.**

This submodule unit tests the private :mod:`beartype.vale._core._valecore`
submodule.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                             }....................
def test_api_vale_validator_pass() -> None:
    '''
    Test successful usage of the private
    :class:`beartype.vale._core._valecore.BeartypeValidator` class.
    '''

    # Defer test-specific imports.
    from beartype.vale._core._valecore import BeartypeValidator

    # Arbitrary valid data validator.
    not_though_the_soldier_knew = lambda text: bool('Someone had blundered.')

    # Keyword arguments passing arguments describing this validator excluding
    # the "get_repr" argument.
    kwargs_sans_get_repr = dict(
        is_valid=not_though_the_soldier_knew,
        is_valid_code_locals={'yum': not_though_the_soldier_knew},
    )

    # Keyword arguments passing arguments describing this validator including
    # the "get_repr" argument.
    kwargs = kwargs_sans_get_repr.copy()
    kwargs['get_repr'] = lambda: (
        "Is[lambda text: bool('Someone had blundered.')]")

    # Code already prefixed by "(" and suffixed by ")".
    is_valid_code_delimited = "({obj} == 'Was there a man dismayed?')"

    # Code *NOT* already prefixed by "(" and suffixed by ")".
    is_valid_code_undelimited = "{obj} == 'All in the valley of Death'"

    # Assert that a beartype validator preserves delimited code as is.
    validator_delimited = BeartypeValidator(
        is_valid_code=is_valid_code_delimited, **kwargs)
    assert validator_delimited._is_valid_code is is_valid_code_delimited

    # Assert that this validator reports the expected diagnosis.
    validator_delimited_diagnosis = validator_delimited.get_diagnosis(
        obj='Flashed all their sabres bare,',
        indent_level_outer='    ',
        indent_level_inner='',
    )
    assert 'True' in validator_delimited_diagnosis
    assert 'Someone had blundered.' in validator_delimited_diagnosis

    # Assert that a beartype validator delimits undelimited code as well as
    # providing non-default validation diagnoses.
    validator_undelimited = BeartypeValidator(
        is_valid_code=is_valid_code_undelimited, **kwargs)
    assert (
        validator_undelimited._is_valid_code ==
        f'({is_valid_code_undelimited})'
    )

    # Assert that a beartype validator also accepts a string representer.
    validator_repr_str = BeartypeValidator(
        is_valid_code=is_valid_code_delimited,
        get_repr='All that was left of them,',
        **kwargs_sans_get_repr
    )

    # Assert these objects have the expected representations.
    assert 'Someone had blundered.' in repr(validator_delimited)
    assert repr(validator_repr_str) == 'All that was left of them,'


def test_api_vale_validator_fail() -> None:
    '''
    Test unsuccessful usage of the private
    :class:`beartype.vale._core._valecore.BeartypeValidator` class.
    '''

    # Defer test-specific imports.
    from beartype.roar import BeartypeValeSubscriptionException
    from beartype.vale._core._valecore import BeartypeValidator
    from pytest import raises

    # Arbitrary valid data validator.
    into_the_jaws_of_death = lambda text: bool('Into the mouth of hell')

    # Assert that attempting to instantiate the "BeartypeValidator" class with
    # non-string code raises the expected exception.
    with raises(BeartypeValeSubscriptionException):
        BeartypeValidator(
            is_valid=into_the_jaws_of_death,
            is_valid_code=b'Into the jaws of Death,',
            is_valid_code_locals={'yum': into_the_jaws_of_death},
            get_repr=lambda: "Is[lambda text: bool('Into the mouth of hell')]",
        )

    # Assert that attempting to instantiate the "BeartypeValidator" class with
    # empty code raises the expected exception.
    with raises(BeartypeValeSubscriptionException):
        BeartypeValidator(
            is_valid=into_the_jaws_of_death,
            is_valid_code='',
            is_valid_code_locals={'yum': into_the_jaws_of_death},
            get_repr=lambda: "Is[lambda text: bool('Into the mouth of hell')]",
        )

    # Assert that attempting to instantiate the "BeartypeValidator" class with
    # code *NOT* containing the substring "{obj}" raises the expected
    # exception.
    with raises(BeartypeValeSubscriptionException):
        BeartypeValidator(
            is_valid=into_the_jaws_of_death,
            is_valid_code='Came through the jaws of Death,',
            is_valid_code_locals={'yum': into_the_jaws_of_death},
            get_repr=lambda: "Is[lambda text: bool('Into the mouth of hell')]",
        )

    # Assert that attempting to instantiate the "BeartypeValidator" class with
    # valid code and non-dictionary code locals raises the expected exception.
    with raises(BeartypeValeSubscriptionException):
        BeartypeValidator(
            is_valid=into_the_jaws_of_death,
            is_valid_code="{obj} == 'Back from the mouth of hell,'",
            is_valid_code_locals={'yum', into_the_jaws_of_death},
            get_repr=lambda: "Is[lambda text: bool('Into the mouth of hell')]",
        )

    # Keyword arguments passing valid code and non-dictionary code locals.
    kwargs_is_valid = dict(
        is_valid=into_the_jaws_of_death,
        is_valid_code="{obj} == 'Back from the mouth of hell,'",
        is_valid_code_locals={'yum': into_the_jaws_of_death},
    )

    # Assert that attempting to instantiate the "BeartypeValidator" class with
    # valid code and code locals but a representer of an invalid type raises
    # the expected exception.
    with raises(BeartypeValeSubscriptionException):
        BeartypeValidator(
            get_repr=b'All that was left of them,', **kwargs_is_valid)

    # Assert that attempting to instantiate the "BeartypeValidator" class with
    # valid code and code locals but an empty-string representer raises the
    # expected exception.
    with raises(BeartypeValeSubscriptionException):
        BeartypeValidator(get_repr='', **kwargs_is_valid)

    # Assert that attempting to instantiate the "BeartypeValidator" class with
    # valid code and code locals but a C-based representer raises the expected
    # exception.
    with raises(BeartypeValeSubscriptionException):
        BeartypeValidator(get_repr=iter, **kwargs_is_valid)

    # Assert that attempting to instantiate the "BeartypeValidator" class with
    # valid code and code locals but a pure-Python representer accepting one or
    # more parameters raises the expected exception.
    with raises(BeartypeValeSubscriptionException):
        BeartypeValidator(
            get_repr=lambda rode, the, six, hundred:
                'Into the valley of Death',
            **kwargs_is_valid
        )

    #FIXME: Uncomment when inevitably needed again. *sigh*
    # # Keyword arguments passing valid code, non-dictionary code locals, and a
    # # representer.
    # kwargs_is_valid_get_repr = dict(
    #     get_repr='Spread far around and inaccessibly',
    #     **kwargs_is_valid
    # )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`585`-compliant **dual type hint utilities**
(i.e., callables generically applicable to both :pep:`484`- and
:pep:`585`-compliant type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep484585Exception
from beartype.typing import (
    Tuple,
    TypeVar,
    Union,
)
from beartype._data.hint.pep.sign.datapepsigns import (
    HintSignForwardRef,
    HintSignType,
    HintSignUnion,
)
from beartype._data.hint.datahinttyping import Pep484585ForwardRef
from beartype._util.cls.pep.utilpep3119 import (
    die_unless_type_issubclassable,
    die_unless_type_or_types_issubclassable,
)
from beartype._util.hint.pep.proposal.pep484585.utilpep484585arg import (
    get_hint_pep484585_args_1)
from typing import (
    Type as typing_Type,  # <-- intentional to distinguish from "type" below
)

# ....................{ HINTS ~ private                    }....................
_HINT_PEP484585_SUBCLASS_ARGS_1_UNION = Union[
    type, Tuple[type], TypeVar, Pep484585ForwardRef,]
'''
Union of the types of all permissible :pep:`484`- or :pep:`585`-compliant
**subclass type hint arguments** (i.e., PEP-compliant child type hints
subscripting (indexing) a subclass type hint).
'''

# ....................{ GETTERS                            }....................
def get_hint_pep484585_type_superclass(
    hint: object,
    exception_prefix: str,
) -> _HINT_PEP484585_SUBCLASS_ARGS_1_UNION:
    '''
    **Issubclassable superclass(es)** (i.e., class whose metaclass does *not*
    define a ``__subclasscheck__()`` dunder method that raises an exception,
    tuple of such classes, or forward reference to such a class) subscripting
    the passed :pep:`484`- or :pep:`585`-compliant **subclass type hint**
    (i.e., hint constraining objects to subclass that superclass).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : object
        Object to be inspected.
    exception_prefix : str
        Human-readable label prefixing the representation of this object in the
        exception message.

    Returns
    -------
    _HINT_PEP484585_SUBCLASS_ARGS_1_UNION
        Argument subscripting this subclass type hint, guaranteed to be either:

        * An issubclassable class.
        * A tuple of issubclassable classes.
        * A :pep:`484`-compliant forward reference to an issubclassable class
          that typically has yet to be declared (i.e.,
          :class:`typing.ForwardRef` instance).
        * A :pep:`484`-compliant type variable constrained to classes (i.e.,
          :class:`typing.TypeVar` instance).
        * A :pep:`585`-compliant union of two or more issubclassable classes.
        * A :pep:`484`-compliant type variable constrained to classes (i.e.,
          :class:`typing.TypeVar` instance).

    Raises
    ------
    BeartypeDecorHintPep3119Exception
        If this superclass subscripting this type hint is *not*
        **issubclassable** (i.e., class whose metaclass defines a
        ``__subclasscheck__()`` dunder method raising an exception).
    BeartypeDecorHintPep484585Exception
        If this hint is either:

        * Neither a :pep:`484`- nor :pep:`585`-compliant subclass type hint.
        * A :pep:`484`- or :pep:`585`-compliant subclass type hint subscripted
          by one argument that is neither a class, union of classes, nor
          forward reference to a class.
    BeartypeDecorHintPep585Exception
        If this hint is either:

        * A :pep:`585`-compliant subclass type hint subscripted by either:

          * *No* arguments.
          * Two or more arguments.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import (
        get_hint_pep_args,
        get_hint_pep_sign,
        get_hint_pep_sign_or_none,
    )

    # If this is neither a PEP 484- *NOR* PEP 585-compliant subclass type hint,
    # raise an exception.
    if get_hint_pep_sign(hint) is not HintSignType:
        raise BeartypeDecorHintPep484585Exception(
            f'{exception_prefix}{repr(hint)} '
            f'neither PEP 484 nor 585 subclass type hint.'
        )
    # Else, this is a subclass type hint.

    # Superclass subscripting this hint.
    hint_superclass = get_hint_pep484585_args_1(
        hint=hint, exception_prefix=exception_prefix)

    # Sign identifying this superclass.
    hint_superclass_sign = get_hint_pep_sign_or_none(hint_superclass)

    # If this superclass is actually a union of superclasses...
    if hint_superclass_sign is HintSignUnion:
        # Efficiently reduce this superclass to the tuple of superclasses
        # subscripting and thus underlying this union.
        hint_superclass = get_hint_pep_args(hint_superclass)

        # If any item of this tuple is *NOT* an issubclassable class, raise an
        # exception.
        # print(f'hint_superclass union arg: {hint_superclass}')
        die_unless_type_or_types_issubclassable(
            type_or_types=hint_superclass, exception_prefix=exception_prefix)  # type: ignore[arg-type]
    # Else, this superclass is *NOT* a union of superclasses.
    #
    # If this superclass is actually a forward reference to a superclass,
    # silently accept this reference as is. This conditional exists only to
    # avoid raising a subsequent exception.
    elif hint_superclass_sign is HintSignForwardRef:
        pass
    # Else, this superclass is *NOT* a forward reference to a superclass.
    #
    # If this superclass is a class...
    elif isinstance(hint_superclass, type):
        die_unless_type_issubclassable(
            cls=hint_superclass, exception_prefix=exception_prefix)
        # Else, this superclass is issubclassable.
    # Else, this superclass is of an unexpected type. In this case, raise an
    # exception.
    #
    # Note that PEP 585-compliant subclass type hints infrequently trigger this
    # edge case. Although the "typing" module explicitly validates the
    # arguments subscripting PEP 484-compliant type hints, the CPython
    # interpreter applies *NO* such validation to PEP 585-compliant subclass
    # type hints. For example, PEP 585-compliant subclass type hints are
    # subscriptable by the empty tuple, which is technically an argument:
    #     >>> type[()].__args__
    #     ()   # <---- thanks fer nuthin
    else:
        raise BeartypeDecorHintPep484585Exception(
            f'{exception_prefix}PEP 484 or 585 subclass type hint '
            f'{repr(hint)} child type hint {repr(hint_superclass)} neither '
            f'class, union of classes, nor forward reference to class.'
        )

    # Return this superclass.
    return hint_superclass  # type: ignore[return-value]

# ....................{ REDUCERS                           }....................
#FIXME: Unit test us up.
def reduce_hint_pep484585_type(
    hint: object, exception_prefix: str, *args, **kwargs) -> object:
    '''
    Reduce the passed :pep:`484`- or :pep:`585`-compliant **subclass type
    hint** (i.e., hint constraining objects to subclass that superclass) to the
    :class:`type` superclass if that hint is subscripted by an ignorable child
    type hint (e.g., :attr:`typing.Any`, :class:`type`) *or* preserve this hint
    as is otherwise (i.e., if that hint is *not* subscripted by an ignorable
    child type hint).

    This reducer is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : object
        Subclass type hint to be reduced.
    exception_prefix : str
        Human-readable label prefixing the representation of this object in the
        exception message.

    All remaining passed arguments are silently ignored.

    Raises
    ------
    BeartypeDecorHintPep484585Exception
        If this hint is neither a :pep:`484`- nor :pep:`585`-compliant subclass
        type hint.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.utilhinttest import is_hint_ignorable

    # If this hint is the unsubscripted PEP 484-compliant subclass type hint,
    # immediately reduce this hint to the "type" superclass.
    #
    # Note that this is *NOT* merely a nonsensical optimization. The
    # implementation of the unsubscripted PEP 484-compliant subclass type hint
    # significantly differs across Python versions. Under some but *NOT* all
    # supported Python versions (notably, Python 3.7 and 3.8), the "typing"
    # module subversively subscripts this hint by a type variable; under all
    # others, this hint remains unsubscripted. In the latter case, passing this
    # hint to the subsequent get_hint_pep484585_args_1() would erroneously
    # raise an exception.
    if hint is typing_Type:
        return type
    # Else, this hint is *NOT* the unsubscripted PEP 484-compliant subclass
    # type hint.

    # Superclass subscripting this hint.
    #
    # Note that we intentionally do *NOT* call the high-level
    # get_hint_pep484585_type_superclass() getter here, as the
    # validation performed by that function would raise exceptions for
    # various child type hints that are otherwise permissible (e.g.,
    # "typing.Any").
    hint_superclass = get_hint_pep484585_args_1(
        hint=hint, exception_prefix=exception_prefix)

    # If this argument is either...
    if (
        # An ignorable type hint (e.g., "typing.Any") *OR*...
        is_hint_ignorable(hint_superclass) or
        # The "type" superclass, which is effectively ignorable in this
        # context of subclasses, as *ALL* classes necessarily subclass
        # that superclass.
        hint_superclass is type
    ):
        # Reduce this subclass type hint to the "type" superclass.
        hint = type
    # Else, this argument is unignorable and thus irreducible.

    # Return this possibly reduced type hint.
    return hint



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-compliant type hint utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._cave._cavefast import NoneType
from beartype._data.hint.pep.sign.datapepsigncls import HintSign
from beartype._data.hint.pep.sign.datapepsigns import (
    HintSignGeneric,
    HintSignNewType,
    HintSignTypeVar,
)
from beartype._data.hint.pep.sign.datapepsignset import HINT_SIGNS_UNION

# Intentionally import PEP 484-compliant "typing" type hint factories rather
# than possibly PEP 585-compliant "beartype.typing" type hint factories.
from typing import (
    Generic,
    Optional,
    Tuple,
)

# ....................{ HINTS                              }....................
HINT_PEP484_TUPLE_EMPTY = Tuple[()]
'''
:pep:`484`-compliant empty fixed-length tuple type hint.
'''

# ....................{ TESTERS ~ ignorable                }....................
def is_hint_pep484_ignorable_or_none(
    hint: object, hint_sign: HintSign) -> Optional[bool]:
    '''
    ``True`` only if the passed object is a :pep:`484`-compliant **ignorable
    type hint,** ``False`` only if this object is a :pep:`484`-compliant
    unignorable type hint, and ``None`` if this object is *not* `PEP
    484`_-compliant.

    Specifically, this tester function returns ``True`` only if this object is
    a deeply ignorable :pep:`484`-compliant type hint, including:

    * A parametrization of the :class:`typing.Generic` abstract base class (ABC)
      by one or more type variables. As the name implies, this ABC is generic
      and thus fails to impose any meaningful constraints. Since a type variable
      in and of itself also fails to impose any meaningful constraints, these
      parametrizations are safely ignorable in all possible contexts: e.g.,

      .. code-block:: python

         from typing import Generic, TypeVar
         T = TypeVar('T')
         def noop(param_hint_ignorable: Generic[T]) -> T: pass

    * The :func:`NewType` closure factory function passed an ignorable child
      type hint. Unlike most :mod:`typing` constructs, that function does *not*
      cache the objects it returns: e.g.,

      .. code-block:: python

         >>> from typing import NewType
         >>> NewType('TotallyNotAStr', str) is NewType('TotallyNotAStr', str)
         False

      Since this implies every call to ``NewType({same_name}, object)`` returns
      a new closure, the *only* means of ignoring ignorable new type aliases is
      dynamically within this function.
    * The :data:`Optional` or :data:`Union` singleton subscripted by one or
      more ignorable type hints (e.g., ``typing.Union[typing.Any, bool]``).
      Why? Because unions are by definition only as narrow as their widest
      child hint. However, shallowly ignorable type hints are ignorable
      precisely because they are the widest possible hints (e.g.,
      :class:`object`, :attr:`typing.Any`), which are so wide as to constrain
      nothing and convey no meaningful semantics. A union of one or more
      shallowly ignorable child hints is thus the widest possible union,
      which is so wide as to constrain nothing and convey no meaningful
      semantics. Since there exist a countably infinite number of possible
      :data:`Union` subscriptions by one or more ignorable type hints, these
      subscriptions *cannot* be explicitly listed in the
      :data:`HINTS_REPR_IGNORABLE_SHALLOW` frozenset. Instead, these
      subscriptions are dynamically detected by this tester at runtime and thus
      referred to as **deeply ignorable type hints.**

    This tester is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as this tester is only safely callable
    by the memoized parent
    :func:`beartype._util.hint.utilhinttest.is_hint_ignorable` tester.

    Parameters
    ----------
    hint : object
        Type hint to be inspected.
    hint_sign : HintSign
        **Sign** (i.e., arbitrary object uniquely identifying this hint).

    Returns
    -------
    Optional[bool]
        Either:

        * If this object is :pep:`484`-compliant:

          * If this object is a ignorable, ``True``.
          * Else, ``False``.

        * If this object is *not* :pep:`484`-compliant, ``None``.
    '''
    # print(f'Testing PEP 484 hint {repr(hint)} [{repr(hint_sign)}] deep ignorability...')

    #FIXME: Remove this *AFTER* properly supporting type variables. For
    #now, ignoring type variables is required ta at least shallowly support
    #generics parametrized by one or more type variables.

    # For minor efficiency gains, the following tests are intentionally ordered
    # in descending likelihood of a match.
    #
    # If this hint is a PEP 484-compliant type variable, unconditionally return
    # true. Type variables require non-trivial and currently unimplemented
    # decorator support.
    if hint_sign is HintSignTypeVar:
        return True
    # Else, this hint is *NOT* a PEP 484-compliant type variable.
    #
    # If this hint is a PEP 484-compliant union...
    elif hint_sign in HINT_SIGNS_UNION:
        # Avoid circular import dependencies.
        from beartype._util.hint.pep.utilpepget import get_hint_pep_args
        from beartype._util.hint.utilhinttest import is_hint_ignorable

        # Return true only if one or more child hints of this union are
        # recursively ignorable. See the function docstring.
        return any(
            is_hint_ignorable(hint_child)
            for hint_child in get_hint_pep_args(hint)
        )
    # Else, this hint is *NOT* a PEP 484-compliant union.
    #
    # If this hint is a PEP 484-compliant generic...
    elif hint_sign is HintSignGeneric:
        # Avoid circular import dependencies.
        from beartype._util.hint.pep.utilpepget import (
            get_hint_pep_origin_or_none)

        # print(f'Testing generic hint {repr(hint)} deep ignorability...')
        # If this generic is the "typing.Generic" superclass directly
        # parametrized by one or more type variables (e.g.,
        # "typing.Generic[T]"), return true.
        #
        # Note that we intentionally avoid calling the
        # get_hint_pep_origin_type_isinstanceable_or_none() function here, which
        # has been intentionally designed to exclude PEP-compliant type hints
        # originating from "typing" type origins for stability reasons.
        if get_hint_pep_origin_or_none(hint) is Generic:
            # print(f'Testing generic hint {repr(hint)} deep ignorability... True')
            return True
        # Else, this generic is *NOT* the "typing.Generic" superclass directly
        # parametrized by one or more type variables and thus *NOT* an
        # ignorable non-protocol.
        #
        # Note that this condition being false is *NOT* sufficient to declare
        # this hint to be unignorable. Notably, the type origin originating
        # both ignorable and unignorable protocols is "Protocol" rather than
        # "Generic". Ergo, this generic could still be an ignorable protocol.
        # print(f'Testing generic hint {repr(hint)} deep ignorability... False')
    # Else, this hint is *NOT* a PEP 484-compliant generic.
    #
    # If this hint is a PEP 484-compliant new type...
    elif hint_sign is HintSignNewType:
        # Avoid circular import dependencies.
        from beartype._util.hint.utilhinttest import is_hint_ignorable
        from beartype._util.hint.pep.proposal.pep484.utilpep484newtype import (
            get_hint_pep484_newtype_alias)

        # Return true only if this hint aliases an ignorable child type hint.
        return is_hint_ignorable(get_hint_pep484_newtype_alias(hint))
    # Else, this hint is *NOT* a PEP 484-compliant new type.

    # Return "None", as this hint is unignorable only under PEP 484.
    return None

# ....................{ REDUCERS                           }....................
#FIXME: Replace the ambiguous parameter:
#* "hint: object" with the unambiguous parameter "hint: Literal[None]" *AFTER*
#  we drop support for Python 3.7. *sigh*

# Note that this reducer is intentionally typed as returning "type" rather than
# "NoneType". While the former would certainly be preferable, mypy erroneously
# emits false positives when this reducer is typed as returning "NoneType":
#     beartype/_util/hint/pep/proposal/pep484/utilpep484.py:190: error: Variable
#     "beartype._cave._cavefast.NoneType" is not valid as a type [valid-type]
def reduce_hint_pep484_none(hint: object, *args, **kwargs) -> type:
    '''
    Reduce the passed :pep:`484`-compliant :data:`None` type hint to the type of
    that type hint (i.e., the builtin :class:`types.NoneType` class).

    While *not* explicitly defined by the :mod:`typing` module, :pep:`484`
    explicitly supports this singleton:

        When used in a type hint, the expression :data:`None` is considered
        equivalent to ``type(None)``.

    This reducer is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : object
        Type variable to be reduced.

    All remaining passed arguments are silently ignored.

    Returns
    -------
    NoneType
        Type of the :data:`None` singleton.
    '''
    assert hint is None, f'Type hint {hint} not "None" singleton.'

    # Unconditionally return the type of the "None" singleton.
    return NoneType



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype import path hook trie** (i.e., data structure caching package names
on behalf of the higher-level :func:`beartype.claw._clawmain` submodule, which
beartype import path hooks internally created by that submodule subsequently
lookup when deciding whether or not (and how) to decorate by
:func:`beartype.beartype` the currently imported user-specific submodule).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.claw._importlib.clawimppath import remove_beartype_pathhook
from beartype.roar import BeartypeClawHookException
from beartype.typing import (
    Dict,
    Iterable,
    # Iterator,
    Optional,
)
from beartype._cave._cavemap import NoneTypeOr
from beartype._conf.confcls import BeartypeConf
# from pprint import pformat

# ....................{ CLASSES                            }....................
#FIXME: Unit test us up, please.
class PackagesTrie(
    #FIXME: Use "beartype.typing.Self" here instead once we backport that.
    Dict[str, Optional['PackagesTrie']]):
    '''
    **(Sub)package configuration (sub)trie** (i.e., recursively nested
    dictionary mapping from the unqualified basename of each subpackage of the
    current package to be runtime type-checked on the first importation of that
    subpackage to another instance of this class similarly describing the
    sub-subpackages of that subpackage).

    This (sub)cache is suitable for caching as the values of:

    * The :data:`.packages_trie` global dictionary.
    * Each (sub)value mapped to by that global dictionary.

    Motivation
    ----------
    This dictionary is intentionally implemented as a nested trie data structure
    rather than a trivial non-nested flat dictionary. Why? Efficiency. Consider
    this flattened set of package names:

        .. code-block:: python

           package_names = {'a.b', 'a.c', 'd'}

    Deciding whether an arbitrary package name is in this set requires
    worst-case ``O(n)`` iteration across the set of ``n`` package names.

    Consider instead this nested dictionary whose keys are package names split
    on ``"."`` delimiters and whose values are either recursively nested
    dictionaries of the same format *or* the :data:`None` singleton (terminating
    the current package name):

        .. code-block:: python

           package_names_trie = {'a': {'b': None, 'c': None}, 'd': None}

    Deciding whether an arbitrary package name is in this dictionary only
    requires worst-case ``O(h)`` iteration across the height ``h`` of this
    dictionary (equivalent to the largest number of ``"."`` delimiters for any
    fully-qualified package name encapsulated by this dictionary). ``h <<<< n``,
    so this dictionary offers *much* faster worst-case lookup than that set.

    Moreover, in the worst case:

    * That set requires one inefficient string prefix test for each item.
    * This dictionary requires *only* one efficient string equality test for
      each nested key-value pair while descending towards the target package
      name.

    Let's do this, fam.

    Caveats
    -------
    **This dictionary is only safely accessible in a thread-safe manner from
    within a** ``with claw_lock:`` **context manager.** Equivalently, this
    dictionary is *not* safely accessible outside that manager.

    Examples
    --------
    An example instance of this dictionary hooked on submodules of the root
    ``package_z`` package, the child ``package_a.subpackage_k`` submodule, and
    the ``package_a.subpackage_b.subpackage_c`` and
    ``package_a.subpackage_b.subpackage_d`` submodules:

        >>> packages_trie = PackagesTrie({
        ...     'package_a': PackagesTrie({
        ...         'subpackage_b': PackagesTrie({
        ...             'subpackage_c': None,
        ...             'subpackage_d': None,
        ...         }),
        ...         'subpackage_k': None,
        ...     }),
        ...     'package_z': None,
        ... })

    Attributes
    ----------
    conf_if_hooked : Optional[BeartypeConf]
        Either:

        * If this (sub)package has been explicitly registered by a prior call to
          the :func:`add_package_names` function, the **beartype
          configuration** (i.e., dataclass encapsulating all settings
          configuring type-checking for this (sub)package).
        * Else, :data:`None`.
    package_basename : Optional[str]
        Either:

        * If this (sub)trie is the global trie :data:`.packages_trie`,
          :data:`None`.
        * Else, the unqualified basename of the (sub)package configured by this
          (sub)trie.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Subclasses declaring uniquely subclass-specific instance
    # variables *MUST* additionally slot those variables. Subclasses violating
    # this constraint will be usable but unslotted, which defeats our purposes.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently called
    # cache dunder methods. Slotting has been shown to reduce read and write
    # costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        'package_basename',
        'conf_if_hooked',
    )

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,
        package_basename : Optional[str],
        *args, **kwargs
    ) -> None:
        '''
        Initialize this package name (sub)cache.

        Parameters
        ----------
        basename : Optional[str]
            Either:

            * If this (sub)trie is the global trie :data:`.packages_trie`,
              :data:`None`.
            * Else, the unqualified basename of the (sub)package configured by
              this (sub)trie.

        All remaining passed parameters are passed as is to the superclass
        :meth:`dict.__init__` method.
        '''
        assert isinstance(package_basename, NoneTypeOr[str]), (
            f'{repr(package_basename)} neither string nor "None".')

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Classify all remaining passed parameters.
        self.package_basename = package_basename

        # Nullify all subclass-specific parameters for safety.
        self.conf_if_hooked: Optional[BeartypeConf] = None

    # ..................{ DUNDERS                            }..................
    def __repr__(self) -> str:

        return '\n'.join((
            f'{self.__class__.__name__}(',
            f'    package_basename={repr(self.package_basename)},',
            f'    conf_if_hooked={repr(self.conf_if_hooked)},',
            f'    dict={super().__repr__()},',
            f')',
        ))

# ....................{ RAISERS                            }....................
def die_if_packages_trie() -> None:
    '''
    Raise an exception if one or more packages have been registered by a prior
    call to the :func:`beartype.claw._pkg.clawpkghook.hook_packages` function.

    Raises
    ------
    BeartypeClawHookException
        If one or more packages have been registered by a prior call to the
        :func:`beartype.claw._pkg.clawpkghook.hook_packages` function.
    '''

    # If one or more packages have been registered...
    if is_packages_trie():
        # Avoid circular import dependencies.
        from beartype.claw._clawstate import claw_state

        # If a global configuration has been added by a prior call to the public
        # beartype.claw.beartype_all() function, raise an exception.
        if claw_state.packages_trie.conf_if_hooked is not None:
            raise BeartypeClawHookException(
                f'Prior call to package-agnostic import hook '
                f'beartype.claw.beartype_all() already registered all packages '
                f'for type-checking under global beartype configuration '
                f'{repr(claw_state.packages_trie.conf_if_hooked)}.'
            )
        # Else, or more package-specific configurations have been added by prior
        # calls to public beartype.claw.beartype_*() functions. In this case,
        # raise another exception.
        else:
            raise BeartypeClawHookException(
                f'Prior call to package-specific import hook '
                f'beartype.claw.beartype_*() already registered some packages '
                f'for type-checking under beartype configurations:\n\t'
                f'{repr(claw_state.packages_trie)}'
            )

# ....................{ TESTERS                            }....................
#FIXME: Unit test us up, please.
def is_packages_trie() -> bool:
    '''
    :data:`True` only if one or more packages have been registered by a prior
    call to the :func:`beartype.claw._pkg.clawpkghook.hook_packages` function.

    Returns
    -------
    bool
        :data:`True` only if one or more packages have been registered.
    '''

    # Avoid circular import dependencies.
    from beartype.claw._clawstate import claw_state

    # Return true only if either...
    return (
        # A global configuration has been added by a prior call to the public
        # beartype.claw.beartype_all() function *OR*...
        claw_state.packages_trie.conf_if_hooked is not None or
        # One or more package-specific configurations have been added by prior
        # calls to public beartype.claw.beartype_*() functions.
        bool(claw_state.packages_trie)
    )

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please.
def get_package_conf_or_none(package_name: str) -> Optional[BeartypeConf]:
    '''
    Beartype configuration with which to type-check the package with the passed
    name if that package *or* a parent package of that package was registered by
    a prior call to the :func:`.hook_packages` function *or* :data:`None`
    otherwise (i.e., if neither that package *nor* a parent package of that
    package was registered by such a call).

    Parameters
    ----------
    package_name : str
        Fully-qualified name of the package to be inspected.

    Returns
    -------
    Optional[BeartypeConf]
        Either:

        * If that package or a parent package of that package was registered by
          a prior call to the :func:`.hook_packages` function, the beartype
          configuration with which to type-check that package.
        * Else, :data:`None`.
    '''

    # Avoid circular import dependencies.
    from beartype.claw._clawstate import claw_state

    # Beartype configuration registered for the currently iterated package,
    # defaulting to the beartype configuration registered for the global trie
    # applicable to *ALL* packages if an external caller previously called the
    # public beartype.claw.beartype_all() function *OR* "None" otherwise (i.e.,
    # if that function has yet to be called).
    subpackage_conf = claw_state.packages_trie.conf_if_hooked

    # For each subpackages trie describing each parent package transitively
    # containing this package (as well as that of that package itself)...
    for subpackages_trie in iter_packages_trie(package_name):
        # Beartype configuration registered with either...
        subpackage_conf = (
            # That parent package if any *OR*...
            #
            # Since that parent package is more granular (i.e., unique) than
            # any transitive parent package of that parent package, the
            # former takes precedence over the latter when defined.
            subpackages_trie.conf_if_hooked or
            # A transitive parent package of that parent package if any.
            subpackage_conf
        )

    # Return this beartype configuration if any *OR* "None" otherwise.
    return subpackage_conf

# ....................{ ITERATORS                          }....................
#FIXME: Unit test us up, please.
def iter_packages_trie(package_name: str) -> Iterable[PackagesTrie]:
    '''
    Generator iteratively yielding one **(sub)package configuration (sub)trie**
    (i.e., :class:`PackagesTrie` instance) describing each transitive parent
    package of the package with the passed name if that package *or* a parent
    package of that package was registered by a prior call to the
    :func:`beartype.claw._pkg.clawpkghook..hook_packages` function *or* the
    empty iterable otherwise otherwise (i.e., if neither that package *nor* a
    parent package of that package was registered by such a call).

    Specifically, this generator yields (in order):

    #. The subtrie of that trie configuring the root package of the passed
       (sub)package.
    #. And so on, until eventually yielding...
    #. The subsubtrie of that subtrie configuring the passed (sub)package
       itself.

    This generator intentionally avoids yielding the global trie
    :data:`.packages_trie`, which is already accessible via that global.

    Parameters
    ----------
    package_name : str
        Fully-qualified name of the package to be inspected.

    Yields
    ------
    PackagesTrie
        (Sub)package configuration (sub)trie describing the currently iterated
        transitive parent package of the package with this name.
    '''
    assert isinstance(package_name, str), f'{repr(package_name)} not string.'

    # Avoid circular import dependencies.
    from beartype.claw._clawstate import (
        claw_lock,
        claw_state,
    )

    # List of each unqualified basename comprising this name, split from this
    # fully-qualified name on "." delimiters. Note that the "str.split('.')" and
    # "str.rsplit('.')" calls produce the exact same lists under all possible
    # edge cases. We arbitrarily call the former rather than the latter for
    # simplicity and readability.
    package_basenames = package_name.split('.')

    # With a submodule-specific thread-safe reentrant lock...
    with claw_lock:
        # Current subtrie of the global trie describing the currently iterated
        # basename of this package, initialized to this global trie itself.
        subpackages_trie: Optional[PackagesTrie] = claw_state.packages_trie

        # For each unqualified basename of each parent package transitively
        # containing this package (as well as that of that package itself)...
        for package_basename in package_basenames:
            # Current subtrie of that trie describing that parent package if
            # that parent package was registered by a prior call to the
            # hook_packages() function *OR* "None" otherwise (i.e., if that
            # parent package has yet to be registered).
            subpackages_trie = subpackages_trie.get(package_basename)  # type: ignore[union-attr]

            # If that parent package has yet to be registered, halt iteration.
            if subpackages_trie is None:
                break
            # Else, that parent package was previously registered.

            # Yield the current subtrie describing that parent package.
            yield subpackages_trie

# ....................{ REMOVERS                           }....................
#FIXME: Unit test us up, please.
def remove_beartype_pathhook_unless_packages_trie() -> None:
    '''
    Remove our **beartype import path hook singleton** (i.e., single callable
    guaranteed to be inserted at most once to the front of the standard
    :mod:`sys.path_hooks` list recursively applying the
    :func:`beartype.beartype` decorator to all well-typed callables and classes
    defined by all submodules of all packages previously registered by a call to
    a public :func:`beartype.claw` function) if this path hook has already been
    added and all previously registered packages have been unregistered *or*
    silently reduce to a noop otherwise (i.e., if either this path hook has yet
    to be added or one or more packages are still registered).

    Caveats
    -------
    **This function is non-thread-safe.** For both simplicity and efficiency,
    the caller is expected to provide thread-safety through a higher-level
    locking primitive managed by the caller.
    '''

    # If all previously registered packages have been unregistered, safely
    # remove our import path hook from the "sys.path_hooks" list.
    if not is_packages_trie():
        remove_beartype_pathhook()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype installer.**

This submodule conforms to the standard :mod:`setuptools`-based "makefile"
format, instrumenting most high-level installation tasks for this package.
'''

# ....................{ TODO                               }....................
#FIXME: Strongly consider migrating to Hatch:
#    https://hatch.pypa.io
#We can't *STAND* poetry, but Hatch looks to be another breed entirely.
#Crucially, PyPA itself has officially adopted Hatch (which is a huge boost),
#Hatch supports dynamic retrieval of version specifiers from Python modules,
#support for PEPs denigrated by poetry authors, and probably much more.
#Basically, Hatch looks like everything we wish poetry was.

# ....................{ KLUDGES ~ path                     }....................
# Explicitly register all files and subdirectories of the root directory
# containing this top-level "setup.py" script to be importable modules and
# packages (respectively) for the remainder of this Python process if this
# directory has yet to be registered.
#
# Technically, this should *NOT* be required. The current build framework
# (e.g., "pip", "setuptools") should implicitly guarantee this to be the case.
# Indeed, the "setuptools"-based "easy_install" script does just that.
# Unfortunately, "pip" >= 19.0.0 does *NOT* guarantee this to be the case for
# projects defining a "pyproject.toml" file -- which, increasingly, is all of
# them. Although "pip" purports to have resolved this upstream, current stable
# releases appear to suffer the same deficiencies. See also:
#     https://github.com/pypa/pip/issues/6163

# Isolate this kludge to a private function for safety.
def _register_dir() -> None:

    # Avert thy eyes, purist Pythonistas!
    import os, sys

    # Absolute dirname of this directory inspired by this StackOverflow answer:
    #     https://stackoverflow.com/a/8663557/2809027
    setup_dirname = os.path.dirname(os.path.realpath(__file__))

    # If the current PYTHONPATH does *NOT* already contain this directory...
    if setup_dirname not in sys.path:
        # Print this registration.
        print(
            'WARNING: Registering "setup.py" directory for importation under '
            'broken installer (e.g., pip >= 19.0.0)...',
            file=sys.stderr)
        # print('setup_dirname: {}\nsys.path: {!r}'.format(setup_dirname, sys.path))

        # Append this directory to the current PYTHONPATH.
        sys.path.append(setup_dirname)

# Kludge us up the bomb.
_register_dir()

# ....................{ KLUDGES ~ init                     }....................
# Explicitly notify the "beartype.__init__" submodule that it is being imported
# at install time from this script. Doing so prevents that submodule from
# implicitly importing from *ANY* "beartype" submodule other than the
# "beartype.meta" submodule, which is the *ONLY* "beartype" submodule guaranteed
# by sheer force of will to be safely importable at install time. All other
# "beartype" submodules should be assumed to be unsafe to import below due to
# potentially importing one or more optional runtime dependencies yet to be
# installed (e.g., the third-party "typing_extensions" package).
#
# Naturally, there exist a countably infinite number of ways to notify the
# "beartype.__init__" submodule that it is being imported at install time. To
# minimize the likelihood of a conflict with other Python subsystems or
# interpreters, we intentionally do *NOT*:
# * Monkey-patch a module or package in the standard library.
# * Detect this script in a stack frame on the call stack.
#
# Instead, we dynamically populate the standard "sys.modules" list of all
# previously imported modules with a fake beartype-specific module. The
# "beartype.__init__" submodule then decides whether to implicitly import from
# potentially unsafe "beartype" submodules by detecting that fake module.

# Isolate this kludge to a private function for safety.
def _notify_beartype() -> None:

    # The acid: it burns more than I expected.
    from sys import modules
    from types import ModuleType

    # Dynamically create a new fake module. Look. Just do it.
    fake_module = ModuleType('beartype.__is_installing__', 'This is horrible.')

    # Dynamically register the fake module. Don't look like that.
    modules['beartype.__is_installing__'] = fake_module

# Setuptools made us do it.
_notify_beartype()

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from packages guaranteed to exist at the start of
# installation. This includes all standard Python and package submodules but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in the installation.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
import setuptools
from beartype import meta

# ....................{ METADATA ~ seo                     }....................
_KEYWORDS = [
    'type checking',
    'type hints',
    'PEP 483',
    'PEP 484',
    'PEP 544',
    'PEP 563',
    'PEP 585',
    'PEP 586',
    'PEP 589',
    'PEP 593',
    'PEP 604',
    'PEP 3141',
]
'''
List of all lowercase alphabetic keywords synopsising this package.

These keywords may be arbitrarily selected so as to pretend to improve search
engine optimization (SEO). In actuality, they do absolutely nothing.
'''

# ....................{ METADATA ~ seo : classifiers       }....................
# To minimize desynchronization woes, all
# "Programming Language :: Python :: "-prefixed strings are dynamically
# appended to this list by the init() function below.
_CLASSIFIERS = [
    # PyPI-specific version type. The number specified here is a magic constant
    # with no relation to this package's version numbering scheme. *sigh*
    'Development Status :: 5 - Production/Stable',

    # Miscellaneous metadata.
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3 :: Only',
    'Topic :: Software Development :: Code Generators',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Quality Assurance',
    'Typing :: Typed',
]
'''
List of all PyPI-specific trove classifier strings synopsizing this project.

Each such string *must* contain either two or three ``" :: "`` substrings
delimiting human-readable capitalized English words formally recognized by the
:mod:`distutils`-specific ``register`` command.

See Also
----------
https://pypi.org/classifiers
    Plaintext list of all trove classifier strings recognized by PyPI.
'''


def _sanitize_classifiers(
    python_version_min_parts: tuple,
    python_version_minor_max: int,
) -> list:
    '''
    List of all PyPI-specific trove classifier strings synopsizing this
    package, manufactured by appending classifiers synopsizing this
    package's support for Python major versions (e.g.,
    ``Programming Language :: Python :: 3.6``, a classifier implying this
    package to successfully run under Python 3.6) to the global list
    :data:`_CLASSIFIERS` of static classifiers.

    Parameters
    ----------
    python_version_min_parts : tuple
        Minimum fully-specified version of Python required by this package
        as a tuple of integers (e.g., ``(3, 5, 0)`` if this package
        requires at least Python 3.5.0).
    python_version_minor_max : int
        Maximum minor stable version of the current Python 3.x mainline (e.g.,
        ``9`` if Python 3.9 is the most recent stable version of Python 3.x).

    Returns
    ----------
    list
        List of all sanitized PyPI-specific trove classifier strings.
    '''
    assert isinstance(python_version_min_parts, tuple), (
        f'"{python_version_min_parts}" not tuple.')
    assert isinstance(python_version_minor_max, int), (
        f'"{python_version_minor_max}" not integer.')

    # Major version of Python required by this package.
    PYTHON_VERSION_MAJOR = python_version_min_parts[0]

    # List of classifiers to return, copied from the global list for safety.
    classifiers_sane = _CLASSIFIERS[:]

    # For each minor version of Python 3.x supported by this package,
    # formally classify this version as such.
    for python_version_minor in range(
        python_version_min_parts[1], python_version_minor_max + 1):
        classifiers_sane.append(
            f'Programming Language :: Python :: '
            f'{PYTHON_VERSION_MAJOR}.{python_version_minor}'
        )
    # print('classifiers: {}'.format(_CLASSIFIERS))

    # Return this sanitized list of classifiers.
    return classifiers_sane

# ....................{ OPTIONS                            }....................
# Setuptools-specific options. Keywords not explicitly recognized by either
# setuptools or distutils must be added to the above dictionary instead.
_SETUP_OPTIONS = {
    # ..................{ CORE                               }..................
    # Self-explanatory metadata. Note that the following metadata keys are
    # instead specified by the "setup.cfg" file:
    #
    # * "license_file", for unknown reasons. We should probably reconsider.
    # * "long_description", since "setup.cfg" supports convenient
    #   "file: ${relative_filename}" syntax for transcluding the contents of
    #   arbitrary project-relative files into metadata values. Attempting to do
    #   so here would require safely opening this file with a context manager,
    #   reading the contents of this file into a local variable, and passing
    #   that variable's value as this metadata outside of that context. (Ugh.)
    'name':             meta.PACKAGE_NAME,
    'version':          meta.VERSION,
    'description':      meta.SYNOPSIS,

    # ..................{ AUTHORS                            }..................
    'author':           meta.AUTHORS,
    'author_email':     meta.AUTHOR_EMAIL,
    'maintainer':       meta.AUTHORS,
    'maintainer_email': meta.AUTHOR_EMAIL,

    # ..................{ URLS                               }..................
    'url':              meta.URL_HOMEPAGE,
    'download_url':     meta.URL_DOWNLOAD,

    # Dictionary mapping from arbitrary human-readable terse names describing
    # various package-related URLs to those URLs.
    'project_urls': {
        'Documentation': meta.URL_HOMEPAGE,
        'Source':        meta.URL_REPO,
        'Issues':        meta.URL_ISSUES,
        'Forums':        meta.URL_FORUMS,
        'Releases':      meta.URL_RELEASES,
    },

    # ..................{ PYPI                               }..................
    # PyPi-specific meta.
    'classifiers': _sanitize_classifiers(
        python_version_min_parts=meta.PYTHON_VERSION_MIN_PARTS,
        python_version_minor_max=meta.PYTHON_VERSION_MINOR_MAX,
    ),
    'keywords': _KEYWORDS,
    'license': meta.LICENSE,

    # ..................{ DEPENDENCIES                       }..................
    # Python dependency.
    'python_requires': f'>={meta.PYTHON_VERSION_MIN}',

    # Mandatory runtime dependencies. This package intentionally requires no
    # such dependencies and hopefully never will.
    'install_requires': (),

    # Optional runtime dependencies. Whereas mandatory dependencies are defined
    # as sequences, optional dependencies are defined as a dictionary mapping
    # from an arbitrary alphanumeric word to a sequence containing one or more
    # such dependencies. Such dependencies are then installable via "pip" by
    # suffixing the name of this project by the "["- and "]"-delimited key
    # defined below whose value lists the dependencies to be installed (e.g.,
    # "sudo pip3 install betse[all]", installing both the package and all
    # mandatory and optional dependencies required by the package).
    'extras_require': {
        # All optional runtime dependencies.
        'all': meta.LIBS_RUNTIME_OPTIONAL,

        # All mandatory developer dependencies (including all mandatory test-
        # and documentation build-time dependencies) as referenced from
        # external project documentation for developers.
        'dev': meta.LIBS_DEVELOPER_MANDATORY,

        # All mandatory Read The Docs (RTD)-specific documentation build-time
        # dependencies an arbitrarily named extra. This is required *ONLY*
        # for integration with the top-level ".readthedocs.yml" file. See the
        # "python" key in that file for further details.
        'doc-rtd': meta.LIBS_DOCTIME_MANDATORY_RTD,

        # All mandatory tox-specific testing dependencies, copied from the
        # "tests_require" key below into an arbitrarily named extra. This is
        # required *ONLY* for integration with the top-level "tox.ini" file.
        # See the "extras" key in that file for further details.
        'test-tox': meta.LIBS_TESTTIME_MANDATORY_TOX,

        # All mandatory coverage-specific testing dependencies as an
        # arbitrarily named extra, required *ONLY* for integration with the
        # top-level "tox.ini" file. See the "extras" key in that file.
        'test-tox-coverage': meta.LIBS_TESTTIME_MANDATORY_COVERAGE,
    },

    # Mandatory testing dependencies.
    'tests_require': meta.LIBS_TESTTIME_MANDATORY_TOX,

    # ..................{ PACKAGES                           }..................
    # List of the fully-qualified names of all Python packages (i.e.,
    # directories containing zero or more Python modules) to be installed,
    # including the top-level package and all subpackages of that package. This
    # thus excludes:
    #
    # * The top-level test package and all subpackages of that package,
    #   defining only test functionality *NOT* intended to be installed with
    #   this package.
    # * "build", caching both setuptools metadata and a complete copy of this
    #   package, required only by a prior package installation.
    #
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # WARNING: This inspection intentionally omits subdirectories containing no
    # "__init__.py" file, despite the remainder of the Python ecosystem
    # commonly accepting such subdirectories as subpackages.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    'packages': setuptools.find_packages(exclude=(
        f'{meta.PACKAGE_NAME}_test',
        f'{meta.PACKAGE_NAME}_test.*',
        'build',
    )),

    # ..................{ PACKAGES ~ data                    }..................
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: *ALL DATA FILES TO BE INSTALLED MUST BE EXPLICITLY MATCHED IN
    # THE TOP-LEVEL "MANIFEST.in" FILE.*
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    #
    # Install all data files (i.e., non-Python files) embedded in the Python
    # package tree for this project which are also explicitly matched by the
    # top-level "MANIFEST.in" file.
    #
    # Unlike Python packages, undesirable data files are includable and
    # excludable from installation *ONLY* via the external "MANIFEST.in" file.
    # This is terrible, of course. (Did you expect otherwise?)
    #
    # Data files are *NOT* Python modules and hence should *NOT* be embedded in
    # the Python package tree. Sadly, the "data_files" key supported by
    # setuptools for this purpose is *NOT* cross-platform-portable and is thus
    # inherently broken. Why? Because that key either requires usage of
    # absolute paths *OR* relative paths relative to absolute paths defined by
    # "setup.cfg"; in either case, those paths are absolute. While the current
    # platform could be detected and the corresponding absolute path embedded
    # in 'data_files', that implementation would be inherently fragile. (That's
    # bad.) In lieu of sane setuptools support, we defer to the methodology
    # employed by everyone. Setuptools, your death is coming.
    #
    # See also:
    # * "Data Files Support", official documentation for this abomination at:
    #   https://setuptools.readthedocs.io/en/latest/userguide/datafiles.html
    'include_package_data': True,

    # Install to an uncompressed directory rather than a compressed archive.
    #
    # While nothing technically precludes the latter, doing so substantially
    # complicates runtime access of data files compressed into this archive
    # (e.g., with the pkg_resources.resource_filename() function). How so? By
    # decompressing this archive's contents into a temporary directory on
    # program startup and removing these contents on program shutdown. Since
    # there exists no guarantee this removal will actually be performed (e.g.,
    # due to preemptive SIGKILLs), compressed archives are inherently fragile.
    #
    # Note that MyPy requires upstream PEP 561-compliant dependencies (like
    # this project) to explicitly prohibit archival. See also:
    #     https://mypy.readthedocs.io/en/stable/installed_packages.html
    'zip_safe': False,
}
'''
Dictionary unpacked as keyword arguments into the subsequent call of the
:func:`setuptools.setup` function, signifying the set of all package-specific
:mod:`setuptools` options.
'''
# print('extras: {}'.format(setup_options['extras_require']))

# ....................{ SETUP                              }....................
setuptools.setup(**_SETUP_OPTIONS)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype caching metaclass unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.cache.utilcachemeta` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_caching_metaclass() -> None:
    '''
    Test the
    :func:`beartype._util.cache.utilcachemeta.BeartypeCachingMeta` metaclass.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableCachedException
    from beartype._util.cache.utilcachemeta import BeartypeCachingMeta
    from pytest import raises

    # ..................{ CALLABLES                          }..................
    class FledNotHisThirstingLips(object, metaclass=BeartypeCachingMeta):
        '''
        Arbitrary class whose metaclass is this metaclass.
        '''

        def __init__(self, and_all_of_great, or_good, or_lovely):
            '''
            Arbitrary callable memoized by this decorator.
            '''

            # If an arbitrary condition, raise an exception whose value depends
            # on these parameters to exercise this metaclass' exception caching.
            if len(or_lovely) == 6:
                raise ValueError(or_lovely)

            # Else, set an instance variable depending on these parameters to
            # exercise this metaclass' instance caching.
            self.which_the_sacred_past = and_all_of_great + or_good + or_lovely

    # ..................{ LOCALS                             }..................
    # Hashable objects to be passed as parameters below.
    and_all_of_great = ('In', 'truth', 'or', 'fable', 'consecrates,', 'he', 'felt',)
    or_good          = ('And', 'knew.', 'When', 'early', 'youth', 'had', 'past,', 'he', 'left',)
    or_lovely        = ('To', 'seek', 'strange', 'truths', 'in', 'undiscovered', 'lands.',)
    he_left          = ('His', 'cold', 'fireside', 'and', 'alienated', 'home',)

    # ..................{ ASSERTS ~ pass                     }..................
    # Assert that memoizing two instances passed the same positional arguments
    # caches and returns the same instance.
    assert (
        FledNotHisThirstingLips(and_all_of_great, or_good, or_lovely) is
        FledNotHisThirstingLips(and_all_of_great, or_good, or_lovely))

    # Assert that memoizing an instance expected to raise an exception does so.
    with raises(ValueError) as exception_first_info:
        FledNotHisThirstingLips(and_all_of_great, or_good, he_left)

    # Assert that repeating that instantiation reraises the same exception.
    with raises(ValueError) as exception_next_info:
        FledNotHisThirstingLips(and_all_of_great, or_good, he_left)
    assert exception_first_info.value is exception_next_info.value

    # Assert that initializing this class with one or more unhashable parameters
    # succeeds with the expected instance.
    and_he_has_bought = FledNotHisThirstingLips(
        ['Many', 'a', 'wide', 'waste',],
        ['and', 'tangled', 'wilderness',],
        ['Has lured his fearless steps;',],
    )
    assert and_he_has_bought.which_the_sacred_past == [
        'Many', 'a', 'wide', 'waste',
        'and', 'tangled', 'wilderness',
        'Has lured his fearless steps;',
    ]
